package fullstack

import (
	"testing"
	"time"

	nats "github.com/nats-io/go-nats"

	xbus "bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/api/control"
	"bitbucket.org/orus-io/go-xbus/tools"
	"bitbucket.org/orus-io/xbus/testutils"
	"github.com/stretchr/testify/assert"
)

func TestPasswordAuth(t *testing.T) {
	ft := InitFullstack(t, true)
	defer ft.Shutdown()

	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	t.Run("Set password", func(t *testing.T) {
		err := ft.Client.Control.Account.PasswordSet(
			control.AccountPasswordSetRequest{
				AccountID: emitter.Client.ID(),
				Password:  "mypass",
			},
		)
		if err != nil {
			t.Error(err)
		}
	})

	t.Run("Password login", func(t *testing.T) {
		options := xbus.COptions{
			Name:        emitter.Client.Name(),
			ConnOptions: ft.Server.NATSConnectOptions(),
		}

		options.ConnOptions.Token = ""
		options.ConnOptions.User = emitter.Client.Name()
		options.ConnOptions.Password = "mypass"

		logger := testutils.GetTestLogger(t)
		client := xbus.NewClient(options, logger)

		err := client.Connect()
		if err != nil {
			t.Error(err)
		}
		assert.Equal(t, emitter.Client.ID(), client.ID())
	})
	t.Run("Invalid password login", func(t *testing.T) {
		ft.Logger.ExpectsError(testutils.ExpectsContains(
			"[server] Account bob: invalid password",
		))
		ft.Logger.ExpectsError(testutils.ExpectsContains(
			"Authorization Error",
		))
		options := xbus.COptions{
			Name:        emitter.Client.Name(),
			ConnOptions: ft.Server.NATSConnectOptions(),
		}

		options.ConnOptions.Token = ""
		options.ConnOptions.User = emitter.Client.Name()
		options.ConnOptions.Password = "wrong"

		logger := testutils.GetTestLogger(t)
		client := xbus.NewClient(options, logger)

		err := client.Connect()
		assert.Equal(t, nats.ErrAuthorization, err)
	})
}

func TestSessionAuth(t *testing.T) {
	ft := InitFullstack(t, true)
	defer ft.Shutdown()

	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	var (
		token  string
		token2 string
	)

	t.Run("Get a session token", func(t *testing.T) {
		if err := ft.Client.Control.Account.PasswordSet(
			control.AccountPasswordSetRequest{
				AccountID: emitter.Client.ID(),
				Password:  "mypass",
			},
		); err != nil {
			t.Error(err)
		}

		options := xbus.COptions{
			Name:        emitter.Client.Name(),
			ConnOptions: ft.Server.NATSConnectOptions(),
		}

		options.ConnOptions.Token = ""
		options.ConnOptions.User = emitter.Client.Name()
		options.ConnOptions.Password = "mypass"

		logger := testutils.GetTestLogger(t)
		client := xbus.NewClient(options, logger)

		if err := client.Connect(); err != nil {
			t.Fatal(err)
		}
		defer client.Close()
		assert.Equal(t, emitter.Client.ID(), client.ID())

		sessionToken, err := client.API.ClientAPI.GetSessionToken(
			api.GetSessionTokenRequest{
				ExpiresIn: 1,
			},
		)
		if err != nil {
			t.Fatal(err)
		}
		token = sessionToken.Token
	})

	t.Run("Authenticate", func(t *testing.T) {
		options := xbus.COptions{
			Name:        emitter.Client.Name(),
			ConnOptions: ft.Server.NATSConnectOptions(),
		}

		options.ConnOptions.Token = "session#" + token

		logger := testutils.GetTestLogger(t)
		client := xbus.NewClient(options, logger)

		if err := client.Connect(); err != nil {
			t.Fatal(err)
		}
		defer client.Close()
		assert.Equal(t, emitter.Client.ID(), client.ID())

	})

	t.Run("Renew token", func(t *testing.T) {
		options := xbus.COptions{
			Name:        emitter.Client.Name(),
			ConnOptions: ft.Server.NATSConnectOptions(),
		}

		options.ConnOptions.Token = "session#" + token

		logger := testutils.GetTestLogger(t)
		client := xbus.NewClient(options, logger)

		if err := client.Connect(); err != nil {
			t.Fatal(err)
		}
		defer client.Close()
		if emitter.Client.ID() != client.ID() {
			t.Fatalf(
				"Not authenticated with the expected account. Expected: %s, got: %s",
				emitter.Client.ID(), client.ID(),
			)
		}

		sessionToken, err := client.API.ClientAPI.GetSessionToken(api.GetSessionTokenRequest{
			InvalidateToken: token,
			ExpiresIn:       1,
		})

		if err != nil {
			t.Fatal(err)
		}

		token2 = sessionToken.Token
	})

	t.Run("Renewed Token", func(t *testing.T) {
		options := xbus.COptions{
			Name:        emitter.Client.Name(),
			ConnOptions: ft.Server.NATSConnectOptions(),
		}

		options.ConnOptions.Token = "session#" + token2

		logger := testutils.GetTestLogger(t)
		client := xbus.NewClient(options, logger)

		if err := client.Connect(); err != nil {
			t.Fatal(err)
		}
		defer client.Close()
		assert.Equal(t, emitter.Client.ID(), client.ID())

	})

	t.Run("Invalidated Token", func(t *testing.T) {
		options := xbus.COptions{
			Name:        emitter.Client.Name(),
			ConnOptions: ft.Server.NATSConnectOptions(),
		}

		options.ConnOptions.Token = "session#" + token

		logger := testutils.GetTestLogger(t)
		client := xbus.NewClient(options, logger)

		ft.Logger.ExpectsError(testutils.ExpectsContains("[server] Unknown session token"))
		ft.Logger.ExpectsError(testutils.ExpectsContains("Authorization Error"))

		if err := client.Connect(); err == nil {
			client.Close()
			t.Fatal("Client should not be able to connect with an invalidated token")
		}
	})

	t.Run("Expired Token", func(t *testing.T) {
		options := xbus.COptions{
			Name:        emitter.Client.Name(),
			ConnOptions: ft.Server.NATSConnectOptions(),
		}

		options.ConnOptions.Token = "session#" + token2

		logger := testutils.GetTestLogger(t)
		client := xbus.NewClient(options, logger)

		time.Sleep(time.Second * 3 / 2)

		ft.Logger.ExpectsError(testutils.ExpectsContains("[server] Session expired"))
		ft.Logger.ExpectsError(testutils.ExpectsContains("Authorization Error"))

		if err := client.Connect(); err == nil {
			client.Close()
			t.Fatal("Client should not be able to connect with an expired token")
		}
	})

	t.Run("Random token", func(t *testing.T) {
		options := xbus.COptions{
			Name:        emitter.Client.Name(),
			ConnOptions: ft.Server.NATSConnectOptions(),
		}

		options.ConnOptions.Token = "session#" + tools.RandomString(32)

		ft.Logger.ExpectsError(testutils.ExpectsContains("[server] Unknown session token"))
		ft.Logger.ExpectsError(testutils.ExpectsContains("Authorization Error"))

		logger := testutils.GetTestLogger(t)
		client := xbus.NewClient(options, logger)

		err := client.Connect()
		if err == nil {
			client.Close()
			t.Fatal("Client should not be able to connect with a random token")
		}
	})
}
