package fullstack

import (
	"context"
	"fmt"
	"testing"
	"time"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"github.com/stretchr/testify/assert"
)

func TestSimpleReply(t *testing.T) {
	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	_, server := ft.CreateWorker("helloworld-actor", []string{"consumerkind1"})

	sub, err := server.Subscribe(func(apc *xbus.ActorProcessingContext) error {
		envelope, err := apc.CompleteEnvelope("request", time.Second)
		if err != nil {
			return err
		}
		sender := apc.OpenOutput("response", "evt1.response")
		if err := sender.AddItems(
			"evt1.response",
			api.Item(fmt.Sprintf("Hello %s", string(envelope.Events[0].Items[0]))),
		); err != nil {
			return err
		}
		if err := sender.Close(); err != nil {
			return err
		}
		return nil
	})
	if err != nil {
		t.Fatal(err)
	}
	defer sub.Unsubscribe()

	assert.Equal(t, "bob", emitter.Name)

	ft.InstallGraph(`
nodes:
- id: source
  type: requestor
  sourcematch:
    eventtypes: [evt1.request]
  outputs: [request]
  inputs: [response]
- id: server
  type: worker
  actors: [helloworld-actor]
  inputs: [request]
  outputs: [response]
edges:
- source.request -> server.request
- server.response -> source.response
`)

	envelope := xbus.ForgeEnvelopeFromItems("evt1.request", []api.Item{api.Item("world")})
	processStateChan, processStateSub, err := emitter.GetClient().
		API.ProcessState.EnvelopeStatesSubscribeChan(
		emitter.ID.String(), envelope.ID.String(),
	)
	if err != nil {
		t.Fatal(err)
	}
	defer processStateSub.Unsubscribe()
	if _, err := emitter.SendEnvelope(nil, "request", envelope); err != nil {
		t.Fatal(err)
	}
	timeout := time.After(4 * time.Second)

waitCompletion:
	for {
		select {
		case <-timeout:
			t.Fatal("Timeout")
		case state := <-processStateChan:
			switch state.Status {
			case api.Process_ERROR:
				t.Fatal("Process error:", state.Errors)
			case api.Process_DONE:
				if state.Response == nil {
					t.Fatal("No response")
				}
				assert.Equal(t, "Hello world",
					string(state.Response.Fragment.Events[0].Items[0]),
				)
				// All good !
				break waitCompletion
			default:
				t.Logf("Process status is %s, waiting for completion waiting", state.Status)
			}
		}
	}

	// Test ProcessState.GetEnvelopeState
	state, err := emitter.GetClient().API.ProcessState.GetEnvelopeState(
		emitter.ID.String(),
		api.GetEnvelopeStateRequest{ID: envelope.ID},
	)

	if err != nil {
		t.Fatal(err)
	}

	assert.Equal(t, api.Process_DONE, state.Status)
	if state.Response == nil {
		t.Fatal("No response")
	}
	assert.Equal(t, "Hello world",
		string(state.Response.Fragment.Events[0].Items[0]),
	)

	var processes []api.Process

	err = ft.Client.ProcessQuery(
		context.Background(), api.ProcessFilter{
			ResultAcked: []bool{true},
		},
		func(p []api.Process) {
			processes = append(processes, p...)
		},
	)
	if err != nil {
		t.Fatal(err)
	}
	if len(processes) != 0 {
		t.Fatalf("Should not have any process, got %v", processes)
	}

	// Acknowledge the result
	if err := emitter.GetClient().API.ProcessState.AckResult(
		emitter.ID.String(),
		api.AckResultRequest{ProcessID: state.ProcessID},
	); err != nil {
		t.Fatal(err)
	}

	err = ft.Client.ProcessQuery(
		context.Background(), api.ProcessFilter{
			ResultAcked: []bool{true},
		},
		func(p []api.Process) {
			processes = append(processes, p...)
		},
	)
	if err != nil {
		t.Fatal(err)
	}

}
