package pipeline

import (
	"errors"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
)

// ErrStatusTransition is when an invalid status transition is requested
var ErrStatusTransition = errors.New("Invalid graph status transition")

// NewManager returns a Manager
func NewManager(
	actorStorage storage.ActorStorage,
	pipelineStorage storage.PipelineStorage,
) *Manager {
	return &Manager{
		ActorStorage:    actorStorage,
		PipelineStorage: pipelineStorage,
	}
}

// Manager manages the pipelines
type Manager struct {
	ActorStorage    storage.ActorStorage
	PipelineStorage storage.PipelineStorage
	Pipelines       []Pipeline
}

func (m *Manager) loadActivePipelines() {
	m.Pipelines = nil

	for _, pipeline := range m.PipelineStorage.LoadActivePipeline() {
		m.Pipelines = append(m.Pipelines, NewPipeline(pipeline))
	}
}

// LookupPipeline returns the matching pipeline for a specific event type
func (m *Manager) LookupPipeline(
	emitter api.Actor, output string, eventtypes []string,
) (*Pipeline, storage.NodeOutputRef, bool) {
	if len(m.Pipelines) == 0 {
		m.loadActivePipelines()
	}
	for _, pipeline := range m.Pipelines {
		output, match := pipeline.Match(emitter, output, eventtypes)
		if match {
			return &pipeline, output, true
		}
	}
	return nil, storage.NodeOutputRef{}, false
}

// GetPipeline load a pipeline from its id
func (m *Manager) GetPipeline(id api.UUID) (Pipeline, bool) {
	pipeline, exists := m.PipelineStorage.Load(id)
	return NewPipeline(pipeline), exists
}

// Save validates and saves a pipeline
func (m *Manager) Save(pipeline *Pipeline) (*Pipeline, error) {
	validator := NewValidator(pipeline.Graph)
	if pipeline.Status != api.PipelineInfo_DRAFT {
		err := validator.Validate()
		if len(err) != 0 {
			return nil, errors.New(
				"Graph is not valid, cannot publish or activate it")
		}
	}
	if pipeline.Status == api.PipelineInfo_ACTIVE {
		err := validator.ValidateActors(m.ActorStorage)
		if err != nil {
			return nil, errors.New(
				"Graph cannot be activated: some roles/actors are unkwown in this bus")
		}
	}
	id, err := m.PipelineStorage.Save(pipeline.Pipeline)
	if err != nil {
		return nil, err
	}
	p, ok := m.GetPipeline(id)
	if !ok {
		panic("Saved pipeline is missing")
	}
	return &p, nil
}

// SetStatus changes a pipeline status
func (m *Manager) SetStatus(pipeline Pipeline, status api.PipelineInfo_Status) error {
	if status == api.PipelineInfo_DRAFT {
		return ErrStatusTransition
	}

	validator := NewValidator(pipeline.Graph)

	if pipeline.Status != api.PipelineInfo_ACTIVE {
		if err := validator.Validate(); err != nil {
			return err
		}

		if err := validator.ValidateActors(m.ActorStorage); err != nil {
			return err
		}
	}

	if err := m.PipelineStorage.SetStatus(pipeline.ID, status); err != nil {
		return err
	}

	// Active pipeline list may have changed, reload it
	m.loadActivePipelines()
	return nil
}

// Query lookup for pipelines matching some criterions
func (m *Manager) Query(
	name string, version string, activeOnly bool,
) (result []Pipeline) {
	for _, pipeline := range m.PipelineStorage.Query(name, version, activeOnly) {
		result = append(result, NewPipeline(pipeline))
	}
	return
}
