package process

import (
	"fmt"

	xbus "bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/pipeline"
	"bitbucket.org/orus-io/xbus/storage"
)

// NewManager creates a process.Manager
func NewManager(
	pipelineManager *pipeline.Manager,
	actorStorage storage.ActorStorage,
	processStorage storage.ProcessStorage,
	logger xbus.Logger,
) *Manager {
	manager := Manager{
		PipelineManager: pipelineManager,
		ActorStorage:    actorStorage,
		ProcessStorage:  processStorage,
		Log:             xbus.NewPrefixedLogger("[process/manager] ", logger),
	}
	return &manager
}

// Manager manage the processes
type Manager struct {
	PipelineManager *pipeline.Manager
	ActorStorage    storage.ActorStorage
	ProcessStorage  storage.ProcessStorage

	Log xbus.Logger

	CurrentProcessList []*Process
}

// Close does nothing for now
func (m *Manager) Close() {
}

func (m *Manager) loadProcess(p storage.Process) (*Process, error) {
	var pipe *pipeline.Pipeline
	if !p.PipelineID.IsZero() {
		p1, exists := m.PipelineManager.GetPipeline(p.PipelineID)
		if !exists {
			return nil, fmt.Errorf("No such pipeline: %s", p.PipelineID)
		}
		pipe = &p1
	}
	return &Process{
		ID:                p.ID,
		CreatedAt:         p.CreatedAt,
		actors:            m.ActorStorage,
		processStorage:    m.ProcessStorage,
		graphManager:      m.PipelineManager,
		manager:           m,
		TriggerEmitterID:  p.TriggerEmitterID,
		TriggerEnvelopeID: p.TriggerEnvelopeID,
		Pipeline:          pipe,
		SourceRef:         p.SourceOutput,
	}, nil
}

// LookupOrCreate returns the Process handling the given envelope or create
// if necessary
func (m *Manager) LookupOrCreate(emitter api.Actor, envelopeID api.UUID) (*Process, error) {
	p, ok := m.ProcessStorage.GetProcessByTrigger(emitter, envelopeID)
	if ok {
		return m.loadProcess(p)
	}

	p = storage.Process{
		TriggerEmitterID:  emitter.ID,
		TriggerEnvelopeID: envelopeID,
	}
	var err error
	p.ID, err = m.ProcessStorage.CreateProcess(p)
	if err != nil {
		return nil, err
	}
	return m.loadProcess(p)
}

// GetEmitterEnvelopeStatus returns the status of the process triggered
// by the given emitter/envelope
func (m *Manager) GetEmitterEnvelopeStatus(emitter api.Actor, envelopeID api.UUID) api.Process_Status {
	p, ok := m.ProcessStorage.GetProcessByTrigger(emitter, envelopeID)
	if !ok {
		return api.Process_NOSTATUS
	}
	return m.ProcessStorage.GetState(p.ID).Status
}

// Get returns a process from its ID
func (m *Manager) Get(id api.UUID) (*Process, error) {
	p, ok := m.ProcessStorage.GetProcess(id)
	if !ok {
		return nil, fmt.Errorf("No such process: %s", id)
	}
	return m.loadProcess(p)
}

// GetByTrigger returns the process matching the given trigger emitter/envelope
func (m *Manager) GetByTrigger(emitter api.Actor, envelopeID api.UUID) (*Process, error) {
	p, ok := m.ProcessStorage.GetProcessByTrigger(emitter, envelopeID)
	if !ok {
		return nil, fmt.Errorf("No process for trigger (%s, %s)", emitter.ID, envelopeID)
	}
	return m.loadProcess(p)

}

// ListPostMortem returns dead processes
func (m *Manager) ListPostMortem(level api.LogLevel, status ...api.PMProcess_Status) ([]*Process, error) {
	var result []*Process

	for _, sProcess := range m.ProcessStorage.QueryPostMortem(level, status...) {
		process, err := m.loadProcess(sProcess)
		if err != nil {
			return nil, err
		}
		result = append(result, process)
	}

	return result, nil
}

// Purge removes any trace of the process in the process storage.
func (m *Manager) Purge(id api.UUID) error {
	// TODO drop the logs
	purged, err := m.ProcessStorage.PurgeProcess([]api.UUID{id})
	if err != nil {
		return err
	}
	if len(purged) != 1 {
		return fmt.Errorf("No terminated process with this id")
	}
	return nil
}

// PurgeMany removes any trace of the processes in the process storage.
func (m *Manager) PurgeMany(ids []api.UUID) ([]api.UUID, error) {
	return m.ProcessStorage.PurgeProcess(ids)
}
