// Copyright 2016 Orus.io. All rights reserved.

package control

import (
	"context"
	"crypto/x509"
	"time"

	nrpc "github.com/nats-rpc/gogo-nrpc"

	"bitbucket.org/orus-io/go-xbus/api/control"
	"bitbucket.org/orus-io/xbus/pipeline"
	"bitbucket.org/orus-io/xbus/process"
	"bitbucket.org/orus-io/xbus/services"
	"bitbucket.org/orus-io/xbus/storage"
)

const (
	queueName = "xbus.control"

	// DefaultPoolSize is the default size of the worker pool
	DefaultPoolSize = 5

	// DefaultPoolMaxPending is the default number of requests that can be
	// pending
	DefaultPoolMaxPending = 5

	// DefaultPoolMaxPendingDuration is the default max pending duration for a
	// a request
	DefaultPoolMaxPendingDuration = time.Second

	// DefaultMaxPurgeJobs is the default maxiumum number of purge jobs
	DefaultMaxPurgeJobs = 1

	// DefaultPurgeBatchSize is the default number of process that are purged
	// in one batch
	DefaultPurgeBatchSize = 200
)

// Options for the Service
type Options struct {
	services.Options

	PoolSize               uint
	PoolMaxPending         uint
	PoolMaxPendingDuration time.Duration

	PurgeBatchSize int

	ServerCert *x509.Certificate
	ServerKey  interface{}

	MaxPurgeJobs int

	PipelineManager *pipeline.Manager
	ProcessManager  *process.Manager
	ProcessStorage  storage.ProcessStorage
	AccountStorage  storage.AccountStorage
	PasswordStorage storage.PasswordStorage
	ActorStorage    storage.ActorStorage
	EnvelopeStorage storage.EnvelopeStorage
	LogStorage      storage.LogStorage
	JobStorage      storage.JobStorage
}

// NewService instanciate a Service
func NewService(o Options) *Service {
	if o.MaxPurgeJobs == 0 {
		o.MaxPurgeJobs = DefaultMaxPurgeJobs
	}
	if o.PoolSize == 0 {
		o.PoolSize = DefaultPoolSize
	}
	if o.PoolMaxPending == 0 {
		o.PoolMaxPending = DefaultPoolMaxPending
	}
	if o.PoolMaxPendingDuration == 0 {
		o.PoolMaxPendingDuration = DefaultPoolMaxPendingDuration
	}
	if o.PurgeBatchSize == 0 {
		o.PurgeBatchSize = DefaultPurgeBatchSize
	}
	return &Service{
		o:           o,
		BaseService: services.NewBaseService("control", o.Options),
	}
}

// Service provides an api for administrators
type Service struct {
	services.BaseService

	workers *nrpc.WorkerPool

	accountHandler   *control.AccountHandler
	actorHandler     *control.ActorHandler
	envelopeHandler  *control.EnvelopeHandler
	pipelineHandler  *control.PipelineHandler
	logsHandler      *control.LogsHandler
	pmprocessHandler *control.PMProcessHandler
	processHandler   *control.ProcessHandler
	o                Options
}

// Startup starts the service
func (s *Service) Startup() error {
	s.workers = nrpc.NewWorkerPool(
		context.Background(),
		s.o.PoolSize, s.o.PoolMaxPending, s.o.PoolMaxPendingDuration)

	s.accountHandler = control.NewAccountHandler(
		context.Background(), s.Conn, accountServer{s},
	)
	s.actorHandler = control.NewActorHandler(
		context.Background(), s.Conn, actorServer{s},
	)
	s.envelopeHandler = control.NewEnvelopeHandler(
		context.Background(), s.Conn, envelopeServer{s},
	)
	s.pipelineHandler = control.NewPipelineHandler(
		context.Background(), s.Conn, pipelineServer{s},
	)
	s.logsHandler = control.NewLogsHandler(
		context.Background(), s.Conn, logsServer{s},
	)
	s.pmprocessHandler = control.NewPMProcessHandler(
		context.Background(), s.Conn, pmprocessServer{s},
	)
	s.processHandler = control.NewProcessConcurrentHandler(
		s.workers, s.Conn, processServer{s},
	)
	if err := s.SubscribeAll(
		s.HandlerQueueSubscribe(s.accountHandler, queueName),
		s.HandlerQueueSubscribe(s.actorHandler, queueName),
		s.HandlerQueueSubscribe(s.envelopeHandler, queueName),
		s.HandlerQueueSubscribe(s.pipelineHandler, queueName),
		s.HandlerQueueSubscribe(s.logsHandler, queueName),
		s.HandlerQueueSubscribe(s.processHandler, queueName),
		s.HandlerQueueSubscribe(s.pmprocessHandler, queueName),
	); err != nil {
		s.Shutdown()
		return err
	}
	return nil
}

// Shutdown stops the workers
func (s *Service) Shutdown() error {
	s.workers.Close(time.Second)
	return nil
}
