// Copyright 2016 Orus.io. All rights reserved.

// Package ravencapture provides the 'core' service, which must be started once and only once
// in every server process
package ravencapture

import (
	"context"
	"fmt"
	"strconv"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/api/control"
	"bitbucket.org/orus-io/xbus/services"
	raven "github.com/getsentry/raven-go"
	yaml "gopkg.in/yaml.v2"
)

func convertMapsKeyToString(m map[interface{}]interface{}) map[string]interface{} {
	var res = make(map[string]interface{})

	for key, value := range m {
		if mv, ok := value.(map[interface{}]interface{}); ok {
			value = convertMapsKeyToString(mv)
		}
		switch k := key.(type) {
		case string:
			res[k] = value
		default:
			res[fmt.Sprintf("%v", k)] = value
		}
	}

	return res
}

func makeJSONReady(value interface{}) interface{} {
	switch value := value.(type) {
	case map[interface{}]interface{}:
		var res = make(map[string]interface{})

		for key, item := range value {
			item = makeJSONReady(item)
			switch k := key.(type) {
			case string:
				res[k] = item
			default:
				res[fmt.Sprintf("%v", k)] = item
			}
		}

		return res
	case []interface{}:
		var res []interface{}
		for _, item := range value {
			item = makeJSONReady(item)
			res = append(res, item)
		}
		return res
	default:
		return value
	}
}

func extractVar(value interface{}, path ...string) interface{} {
	if len(path) == 0 {
		return value
	}
	key := path[0]
	subPath := path[1:]
	switch value := value.(type) {
	case map[string]interface{}:
		return extractVar(value[key], subPath...)
	case map[interface{}]interface{}:
		return extractVar(value[key], subPath...)
	case []interface{}:
		iKey, err := strconv.Atoi(key)
		if err != nil {
			return nil
		}
		return value[iKey]
	default:
		return nil
	}
}

func extractVarString(value interface{}, path ...string) string {
	s, _ := extractVar(value, path...).(string)
	return s
}

// Options is used to create a 'ravencapture' service
type Options struct {
	services.Options

	RavenClient *raven.Client
}

// NewService creates a Emitter
func (o Options) NewService() *Service {
	return &Service{
		BaseService: services.NewBaseService("ravencapture", o.Options),
		o:           o,
	}
}

// Service is in charge of emitting envelopes to Actors
type Service struct {
	services.BaseService

	o Options
}

// Startup starts the service
func (s *Service) Startup() error {
	s.Log.Noticef("Startup")
	api := api.NewClient(s.Conn)
	sub, err := api.CoreEvents.ProcessSubscribe(s.handleCoreEventsProcess)
	if err != nil {
		return err
	}
	s.AddSubscription(sub)
	return nil
}

func (s *Service) handleCoreEventsProcess(evt api.CoreEvent) {
	if evt.Process.Status == api.Process_ERROR {
		c := control.NewProcessClient(s.Conn)

		var buf []byte

		if err := c.ExportStream(
			context.Background(),
			control.ProcessExportRequest{
				ProcessIDs:        []api.UUID{evt.Process.ID},
				ExportEnvelope:    true,
				ExportJobs:        false,
				ExportLogs:        true,
				ExportProcessLogs: true,
				EnvelopeSizeLimit: 1024 * 10,
				LogLevel:          api.LogLevel_ERROR,
			},
			func(ctx context.Context, res control.OctetStreamList) {
				for _, chunk := range res.GetChunks() {
					buf = append(buf, chunk.Data...)
				}
			},
		); err != nil {
			s.Log.Errorf("Could not export process %s: %s", evt.Process.ID, err)
			return
		}

		var data []map[string]interface{}

		if err := yaml.Unmarshal(buf, &data); err != nil {
			s.Log.Errorf("Process export could not be parsed yaml: %s", err)
			return
		}

		if len(data) != 1 {
			s.Log.Errorf("Expected 1 process in the export, got %d", len(data))
			return
		}

		var (
			tags   = make(map[string]string)
			extra  = make(map[string]interface{})
			reason = "N/A"
		)

		tags["xbus.emitter"] = extractVarString(data[0], "trigger", "emitter", "name")

		if value := extractVarString(data[0], "state", "statusreason"); value != "" {
			reason = value
		} else {
			s.Log.Errorf("Could not extract StatusReason")
		}

		if value := extractVarString(data[0], "pipeline", "pipelineinfo", "name"); value != "" {
			tags["xbus.pipeline.name"] = value
		}
		if value := extractVarString(data[0], "pipeline", "pipelineinfo", "version"); value != "" {
			tags["xbus.pipeline.version"] = value
		}

		for name, data := range data[0] {
			data = makeJSONReady(data)
			if name == "id" {
				name = "process_id"
			}
			name = "xbus." + name
			extra[name] = data
		}

		message := raven.Message{
			Message: fmt.Sprintf("Process %%s failed for the following reason: %s", reason),
			Params:  []interface{}{evt.Process.ID},
		}
		packet := raven.NewPacketWithExtra(
			fmt.Sprintf(message.Message, message.Params...),
			extra,
			&message)

		_, ch := s.o.RavenClient.Capture(packet, tags)
		go func() {
			if err := <-ch; err != nil {
				s.Log.Errorf("Sending to sentry failed: %s", err)
			}
		}()
	}
}
