package storage

import (
	"fmt"
	"strings"

	"bitbucket.org/orus-io/go-xbus/api"
)

// NodeType is the type of Node
//go:generate enum-marshaler -type NodeType -suffix Node -lower graphtypes.go
type NodeType int

const (
	// EmitterNode is the type of emitter nodes
	EmitterNode NodeType = iota
	// ConsumerNode is the type of consumer nodes
	ConsumerNode
	// WorkerNode is the type of worker nodes
	WorkerNode
	// RequestorNode is an emitter that waits for a reply
	RequestorNode
)

// SourceMatch are rules to match incoming envelope with
type SourceMatch struct {
	EventTypes []string // All event types envelope must contain exactly
}

// Node is a graph vertex
type Node struct {
	ID            string
	Type          NodeType
	ActorIDs      []api.UUID // Explicit actors. Deprecated (will be removed in a later version)
	Actors        []string   // Explicit actors, by name or UUID
	Roles         []string   // Actor roles
	RoleBroadcast bool       // If true, all actors matching each role will receive the message, instead of only one actor per role

	SourceMatch *SourceMatch // define only if Type==EmitterNode

	Inputs  []string
	Outputs []string

	// TODO add 'arguments'
}

// NodeInputRef is a reference to a node input
type NodeInputRef struct {
	NodeID string
	Input  string
}

// NodeOutputRef is a reference to a node input
type NodeOutputRef struct {
	NodeID string
	Output string
}

// Edge is a Graph oriented edge
type Edge struct {
	From NodeOutputRef
	To   NodeInputRef
}

// Graph stores the graph definition of a pipeline
type Graph struct {
	Nodes []Node
	Edges []Edge
}

func parseNodeIO(str string) (nodeID string, ioName string, err error) {
	for str[0] == ' ' {
		str = str[1:]
	}
	for str[len(str)-1] == ' ' {
		str = str[:len(str)-1]
	}
	splitted := strings.Split(str, ".")
	if len(splitted) > 2 {
		err = fmt.Errorf("Invalid NodeInputRef '%s': expects at most one '.'", str)
		return
	}
	nodeID = splitted[0]

	if len(splitted) == 2 {
		ioName = splitted[1]
	} else {
		ioName = "default"
	}
	return
}

// NodeInputRefFromString loads a NodeInputRef
func NodeInputRefFromString(str string) (ref NodeInputRef, err error) {
	ref.NodeID, ref.Input, err = parseNodeIO(str)
	return
}

// String serialize a NodeInputRef
func (r NodeInputRef) String() string {
	return r.NodeID + "." + r.Input
}

// MarshalYAML marshals NodeInputRef to string
func (r NodeInputRef) MarshalYAML() (interface{}, error) {
	return r.String(), nil
}

// UnmarshalYAML unmarshals NodeInputRef from string
func (r *NodeInputRef) UnmarshalYAML(unmarshal func(interface{}) error) error {
	var str string
	var err error
	if err = unmarshal(&str); err != nil {
		return err
	}
	r.NodeID, r.Input, err = parseNodeIO(str)
	return err
}

// NodeOutputRefFromString loads a NodeInputRef
func NodeOutputRefFromString(str string) (ref NodeOutputRef, err error) {
	ref.NodeID, ref.Output, err = parseNodeIO(str)
	return
}

// String serialize a NodeOutputRef
func (r NodeOutputRef) String() string {
	return r.NodeID + "." + r.Output
}

// MarshalYAML marshals NodeOutputRef to string
func (r NodeOutputRef) MarshalYAML() (interface{}, error) {
	return r.String(), nil
}

// UnmarshalYAML unmarshals NodeOutputRef from string
func (r *NodeOutputRef) UnmarshalYAML(unmarshal func(interface{}) error) error {
	var str string
	var err error
	if err = unmarshal(&str); err != nil {
		return err
	}
	r.NodeID, r.Output, err = parseNodeIO(str)
	return err
}

// EdgeFromString unserialize a Edge
func EdgeFromString(str string) (Edge, error) {
	sp := strings.Split(str, "->")
	if len(sp) != 2 {
		return Edge{}, fmt.Errorf("Invalid edge: %s", str)
	}
	from, err := NodeOutputRefFromString(sp[0])
	if err != nil {
		return Edge{}, fmt.Errorf("Syntax error: output '%s'", str)
	}
	to, err := NodeInputRefFromString(sp[1])
	if err != nil {
		return Edge{}, fmt.Errorf("Syntax error: input: '%s'", str)
	}
	return Edge{from, to}, nil
}

// String() serialize a Edge
func (e Edge) String() string {
	return e.From.String() + "->" + e.To.String()
}

// MarshalYAML marshals Edge to string
func (e Edge) MarshalYAML() (interface{}, error) {
	return e.String(), nil
}

// UnmarshalYAML unmarshals Edge from string
func (e *Edge) UnmarshalYAML(unmarshal func(interface{}) error) error {
	var (
		str string
		err error
	)
	if err = unmarshal(&str); err != nil {
		return err
	}
	*e, err = EdgeFromString(str)
	return err
}
