package sql

import (
	xbus "bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"github.com/m4rw3r/uuid"
	"github.com/orus-io/yago"
)

// NewPasswordStorage returns a PasswordStorage instance
func NewPasswordStorage(db *yago.DB, logger xbus.Logger) *PasswordStorage {
	return &PasswordStorage{RawSQLBase{db, logger}}
}

// PasswordStorage provides a SQL based storage.PasswordStorage implementaion
type PasswordStorage struct {
	RawSQLBase
}

// Set sets a password the a given id
func (s *PasswordStorage) Set(id api.UUID, password string) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	_, err := tx.Exec(
		"INSERT INTO passwords (account_id, password) "+
			"VALUES ($1, crypt($2, gen_salt('bf'))) "+
			"ON CONFLICT (account_id) "+
			"DO UPDATE SET password = crypt($2, gen_salt('bf'))",
		uuid.UUID(id), password,
	)
	if err != nil {
		panic(err)
	}

	tx.MustCommit()
}

// Reset removed any password for the given id
func (s *PasswordStorage) Reset(id api.UUID) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	_, err := tx.Exec(
		"DELETE FROM passwords "+
			"WHERE account_id = $1",
		uuid.UUID(id),
	)
	if err != nil {
		panic(err)
	}

	tx.MustCommit()
}

// Verify returns true if the given password is correct
func (s *PasswordStorage) Verify(id api.UUID, password string) bool {
	tx := s.MustBegin()
	defer tx.Rollback()

	row := tx.QueryRow("SELECT EXISTS("+
		"SELECT 1 "+
		"FROM passwords "+
		"WHERE account_id = $1 "+
		"  AND password = crypt($2, password))",
		uuid.UUID(id), password,
	)

	var correct bool
	if err := row.Scan(&correct); err != nil {
		panic(err)
	}
	return correct
}
