package sql

import (
	"fmt"
	"strings"

	"github.com/m4rw3r/uuid"
	"github.com/orus-io/yago"
	"github.com/slicebit/qb"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
)

func readStringList(s string) []string {
	if s == "" {
		return []string{}
	}
	return strings.Split(s, ",")
}

func writeStringList(list []string) string {
	return strings.Join(list, ",")
}

// NewPipelineStorage initialize a PipelineStorage
func NewPipelineStorage(model *model.Model, db *yago.DB, logger xbus.Logger) *PipelineStorage {
	return &PipelineStorage{Base{
		model:  model,
		db:     db,
		logger: logger,
	}}
}

// PipelineStorage implements storage.PipelineStorage
type PipelineStorage struct {
	Base
}

func dbNodeToStorage(dbNode model.GraphNode) storage.Node {
	node := storage.Node{
		ID:            dbNode.NodeID,
		RoleBroadcast: dbNode.RoleBroadcast,
		Inputs:        readStringList(dbNode.Inputs),
		Outputs:       readStringList(dbNode.Outputs),
	}
	switch dbNode.Type {
	case "emitter":
		node.Type = storage.EmitterNode
		node.SourceMatch = &storage.SourceMatch{
			EventTypes: strings.Split(dbNode.SourceMatchEventTypes, ","),
		}
	case "consumer":
		node.Type = storage.ConsumerNode
	case "worker":
		node.Type = storage.WorkerNode
	case "requestor":
		node.Type = storage.RequestorNode
		node.SourceMatch = &storage.SourceMatch{
			EventTypes: strings.Split(dbNode.SourceMatchEventTypes, ","),
		}
	}
	for _, id := range strings.Split(dbNode.ActorIDs, ",") {
		if id != "" {
			node.ActorIDs = append(node.ActorIDs, api.MustUUIDFromString(id))
		}
	}
	for _, actor := range strings.Split(dbNode.Actors, ",") {
		if actor != "" {
			node.Actors = append(node.Actors, actor)
		}
	}
	if dbNode.Roles != "" {
		node.Roles = strings.Split(dbNode.Roles, ",")
	}
	return node
}

func parseEdgeList(str string) ([]storage.Edge, error) {
	var edges []storage.Edge
	for _, strEdge := range strings.Split(str, ",") {
		if strEdge == "" {
			continue
		}
		edge, err := storage.EdgeFromString(strEdge)
		if err != nil {
			return nil, err
		}
		edges = append(edges, edge)
	}
	return edges, nil
}

func writeEdgeList(edges []storage.Edge) string {
	var strEdges []string
	for _, edge := range edges {
		strEdges = append(strEdges, edge.String())
	}
	return strings.Join(strEdges, ",")
}

func dbPipelineToStorage(dbPipeline model.Pipeline, dbNodes []model.GraphNode) (storage.Pipeline, error) {
	var nodes []storage.Node
	for _, dbNode := range dbNodes {
		nodes = append(nodes, dbNodeToStorage(dbNode))
	}
	edges, err := parseEdgeList(dbPipeline.Edges)
	if err != nil {
		return storage.Pipeline{}, err
	}
	return storage.Pipeline{
		PipelineInfo: api.PipelineInfo{
			ID:      api.UUID(dbPipeline.ID),
			Name:    dbPipeline.Name,
			Version: dbPipeline.Version,
			Status:  dbPipeline.Status,
		},
		Graph: storage.Graph{
			Nodes: nodes,
			Edges: edges,
		},
	}, nil
}

var storageNodeTypeToDB = []string{"emitter", "consumer", "worker", "requestor"}

func storageNodeToDB(node storage.Node) model.GraphNode {
	var strActorIDs []string
	for _, id := range node.ActorIDs {
		strActorIDs = append(strActorIDs, id.String())
	}

	var sourceMatchEventTypes string
	if (node.Type == storage.EmitterNode ||
		node.Type == storage.RequestorNode) &&
		node.SourceMatch != nil {
		sourceMatchEventTypes = strings.Join(node.SourceMatch.EventTypes, ",")
	}

	return model.GraphNode{
		NodeID:                node.ID,
		Type:                  storageNodeTypeToDB[node.Type],
		ActorIDs:              strings.Join(strActorIDs, ","),
		Actors:                strings.Join(node.Actors, ","),
		Roles:                 writeStringList(node.Roles),
		RoleBroadcast:         node.RoleBroadcast,
		SourceMatchEventTypes: sourceMatchEventTypes,
		Inputs:                writeStringList(node.Inputs),
		Outputs:               writeStringList(node.Outputs),
	}
}

func storagePipelineToDB(pipeline storage.Pipeline) (model.Pipeline, []model.GraphNode) {
	var dbNodes []model.GraphNode
	for _, node := range pipeline.Graph.Nodes {
		dbNodes = append(dbNodes, storageNodeToDB(node))
	}
	return model.Pipeline{
		Base:    model.Base{ID: uuid.UUID(pipeline.ID)},
		Name:    pipeline.Name,
		Version: pipeline.Version,
		Status:  pipeline.Status,
		Edges:   writeEdgeList(pipeline.Graph.Edges),
	}, dbNodes
}

// Load loads a graph from storage
func (s *PipelineStorage) Load(id api.UUID) (storage.Pipeline, bool) {
	tx := s.MustBegin()
	defer tx.Rollback()
	var dbPipeline model.Pipeline
	err := tx.Query(s.model.Pipeline).Get(&dbPipeline, uuid.UUID(id))
	if err == yago.ErrRecordNotFound {
		return storage.Pipeline{}, false
	}
	must(err)

	var dbNodes []model.GraphNode
	must(
		tx.Query(s.model.GraphNode).
			Where(s.model.GraphNode.PipelineID.Eq(uuid.UUID(id))).
			All(&dbNodes))

	pipeline, err := dbPipelineToStorage(dbPipeline, dbNodes)
	must(err)

	return pipeline, true
}

// Save saves a pipeline to storage
func (s *PipelineStorage) Save(pipeline storage.Pipeline) (api.UUID, error) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	var dbPipeline model.Pipeline

	fmt.Println(pipeline)
	if !pipeline.ID.IsZero() {
		if err := tx.Query(s.model.Pipeline).
			Where(s.model.Pipeline.ID.Eq(uuid.UUID(pipeline.ID))).
			One(&dbPipeline); err != nil {
			if err == yago.ErrRecordNotFound {
				err = fmt.Errorf("No such pipeline: %s", pipeline.ID)
				return api.UUID{}, err
			}
			must(err)
		}
		if dbPipeline.Status != api.PipelineInfo_DRAFT {
			return api.UUID{}, storage.ErrInvalidPipelineStatus
		}
	} else {
		if err := tx.Query(s.model.Pipeline).
			Where(
				s.model.Pipeline.Name.Eq(pipeline.Name),
				s.model.Pipeline.Version.Eq(pipeline.Version)).
			One(&dbPipeline); err != nil {
			if err != yago.ErrRecordNotFound {
				must(err)
			}
		} else {
			if dbPipeline.Status != api.PipelineInfo_DRAFT {
				return api.UUID{}, storage.ErrInvalidPipelineStatus
			}
		}
	}

	isNewPipeline := dbPipeline.ID.IsZero()

	nextPipeline, dbNodes := storagePipelineToDB(pipeline)

	dbPipeline.Name = nextPipeline.Name
	dbPipeline.Version = nextPipeline.Version
	dbPipeline.Edges = nextPipeline.Edges

	if isNewPipeline {
		s.logger.Noticef("Inserting pipeline")
		must(
			tx.Insert(&dbPipeline))
	} else {
		s.logger.Noticef("Updating pipeline")
		must(
			tx.Update(
				&dbPipeline,
				model.PipelineName,
				model.PipelineVersion,
			))
	}

	s.logger.Noticef("Got %s", dbPipeline.ID, dbNodes)
	var dbNodeIDs []interface{}

	q := tx.Query(s.model.GraphNode).
		Select(s.model.GraphNode.ID).
		Where(s.model.GraphNode.PipelineID.Eq(dbPipeline.ID))

	for _, dbNode := range dbNodes {
		dbNode.PipelineID = dbPipeline.ID
		var id int64
		if err := q.Filter(
			s.model.GraphNode.NodeID.Eq(dbNode.NodeID),
		).Scalar(&id); err != nil {
			if err == yago.ErrRecordNotFound {
				s.logger.Noticef("Insert node %d (%s)", id, dbNode)
				must(
					tx.Insert(&dbNode))
			} else {
				must(err)
			}
		} else {
			dbNode.ID = id
			s.logger.Noticef("Update node %d (%s)", id, dbNode)
			must(
				tx.Update(&dbNode))
		}
		dbNodeIDs = append(dbNodeIDs, dbNode.NodeID)
	}

	_, err := tx.GetEngine().Exec(
		qb.Delete(*s.model.GraphNode.GetMapper().Table()).
			Where(qb.And(
				s.model.GraphNode.PipelineID.Eq(dbPipeline.ID),
				s.model.GraphNode.NodeID.NotIn(dbNodeIDs...))),
	)
	must(err)
	tx.MustCommit()

	return api.UUID(dbPipeline.ID), nil
}

// SetStatus changes a graph status
func (s *PipelineStorage) SetStatus(id api.UUID, status api.PipelineInfo_Status) error {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	if status != api.PipelineInfo_ACTIVE && status != api.PipelineInfo_INACTIVE {
		return storage.ErrInvalidPipelineStatus
	}

	if status == api.PipelineInfo_ACTIVE {
		// Deactivate any other version of the same graph
		var name string
		must(tx.Query(s.model.Pipeline).
			Where(s.model.Pipeline.ID.Eq(uuid.UUID(id))).
			Select(s.model.Pipeline.Name).
			Scalar(&name))

		_, err := tx.GetEngine().Exec(
			s.model.Pipeline.GetMapper().Table().Update().
				Values(map[string]interface{}{
					model.PipelineStatusColumnName: "INACTIVE",
				}).
				Where(qb.And(
					s.model.Pipeline.Name.Eq(name),
					s.model.Pipeline.Status.Eq("ACTIVE"),
				)),
		)
		must(err)
	}

	must(tx.Update(
		&model.Pipeline{Base: model.Base{ID: uuid.UUID(id)}, Status: status},
		model.PipelineStatus))

	tx.MustCommit()

	return nil
}

// LoadActivePipeline loads all the active graphs
func (s *PipelineStorage) LoadActivePipeline() []storage.Pipeline {
	tx := s.MustBegin()
	defer tx.Rollback()

	var dbPipelines []model.Pipeline
	must(
		tx.Query(s.model.Pipeline).
			Where(s.model.Pipeline.Status.Eq("ACTIVE")).
			All(&dbPipelines))

	var graphList []storage.Pipeline

	for _, dbPipeline := range dbPipelines {
		var dbNodes []model.GraphNode
		must(
			tx.Query(s.model.GraphNode).
				Where(s.model.GraphNode.PipelineID.Eq(dbPipeline.ID)).
				All(&dbNodes))
		graph, err := dbPipelineToStorage(dbPipeline, dbNodes)
		must(err)
		graphList = append(graphList, graph)
	}
	return graphList
}

// Query searches for matching graphs
func (s *PipelineStorage) Query(name, version string, activeOnly bool) (graphs []storage.Pipeline) {
	tx := s.MustBegin()
	defer tx.Rollback()

	q := tx.Query(s.model.Pipeline)
	if name != "" {
		q = q.Filter(s.model.Pipeline.Name.Eq(name))
	}
	if version != "" {
		q = q.Filter(s.model.Pipeline.Version.Eq(version))
	}
	if activeOnly {
		q = q.Filter(s.model.Pipeline.Status.Eq(api.PipelineInfo_ACTIVE))
	}

	var dbPipelines []model.Pipeline
	must(q.All(&dbPipelines))

	for _, dbPipeline := range dbPipelines {
		var dbNodes []model.GraphNode
		must(
			tx.Query(s.model.GraphNode).
				Where(s.model.GraphNode.PipelineID.Eq(dbPipeline.ID)).
				All(&dbNodes))

		graph, err := dbPipelineToStorage(dbPipeline, dbNodes)
		must(err)

		graphs = append(graphs, graph)
	}
	return
}
