// Copyright 2016 Orus.io. All rights reserved.

package sql

import (
	"encoding/json"
	"fmt"
	"strconv"
	"time"

	"github.com/lib/pq"
	"github.com/m4rw3r/uuid"
	"github.com/orus-io/yago"
	"github.com/slicebit/qb"

	xbus "bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
)

// NewProcessStorage initialize a new ProcessStorage
func NewProcessStorage(model *model.Model, db *yago.DB, logger xbus.Logger) *ProcessStorage {
	return &ProcessStorage{
		Base: Base{
			model:  model,
			db:     db,
			logger: logger,
		},
		onStatusChanged:        storage.NoOpProcessEventStatusHandler,
		processLogEntryHandler: storage.VoidProcessLogEntryHandler,
	}
}

// ProcessStorage implements storage.ProcessStorage
type ProcessStorage struct {
	Base
	onStatusChanged        storage.ProcessEventStatusHandler
	processLogEntryHandler storage.ProcessLogEntryHandler
}

// SetEventHandlers sets the event handlers
func (s *ProcessStorage) SetEventHandlers(onStatusChanged storage.ProcessEventStatusHandler) {
	if onStatusChanged == nil {
		s.onStatusChanged = storage.NoOpProcessEventStatusHandler
	} else {
		s.onStatusChanged = onStatusChanged
	}
}

// SetProcessLogEntryHandler sets the process logentry handler
func (s *ProcessStorage) SetProcessLogEntryHandler(
	handler storage.ProcessLogEntryHandler,
) {
	if handler == nil {
		s.processLogEntryHandler = storage.VoidProcessLogEntryHandler
	} else {
		s.processLogEntryHandler = handler
	}
}

func dbProcessToStorage(dbProcess model.Process) storage.Process {
	return storage.Process{
		ID:                api.UUID(dbProcess.ID),
		TriggerEmitterID:  api.UUID(dbProcess.TriggerEmitterID),
		TriggerEnvelopeID: api.UUID(dbProcess.TriggerEnvelopeID),
		CreatedAt:         dbProcess.CreatedAt,
		PipelineID:        api.UUID(dbProcess.PipelineID),
		SourceOutput: storage.NodeOutputRef{
			NodeID: dbProcess.SourceNodeID,
			Output: dbProcess.SourceSlotID,
		},
	}
}

func storageProcessToDB(process storage.Process) model.Process {
	return model.Process{
		Base: model.Base{ID: uuid.UUID(process.ID),
			CreatedAt: process.CreatedAt,
		},
		TriggerEmitterID:  uuid.UUID(process.TriggerEmitterID),
		TriggerEnvelopeID: uuid.UUID(process.TriggerEnvelopeID),
		PipelineID:        uuid.UUID(process.PipelineID),
		SourceNodeID:      process.SourceOutput.NodeID,
		SourceSlotID:      process.SourceOutput.Output,
	}
}

func (s *ProcessStorage) loadProcess(tx *yago.Tx, id api.UUID) (storage.Process, bool) {
	var dbProcess model.Process
	if err := tx.Query(s.model.Process).Get(&dbProcess, uuid.UUID(id)); err != nil {
		if err == yago.ErrRecordNotFound {
			return storage.Process{}, false
		}
		must(err)
	}
	return dbProcessToStorage(dbProcess), true
}

// GetProcess returns a process data
func (s *ProcessStorage) GetProcess(id api.UUID) (storage.Process, bool) {
	tx := s.MustBegin()
	defer tx.Rollback()
	return s.loadProcess(tx.Tx, id)
}

// GetProcessByTrigger returns a process matching a trigger or an empty value if not found
// TODO in the future we should support multiple process per trigger.
func (s *ProcessStorage) GetProcessByTrigger(emitter api.Actor, envelopeID api.UUID) (storage.Process, bool) {
	tx := s.MustBegin()
	defer tx.Rollback()
	var dbProcess model.Process
	if err := tx.Query(s.model.Process).Where(
		s.model.Process.TriggerEmitterID.Eq(uuid.UUID(emitter.ID)),
		s.model.Process.TriggerEnvelopeID.Eq(uuid.UUID(envelopeID)),
	).One(&dbProcess); err != nil {
		if err == yago.ErrRecordNotFound {
			return storage.Process{}, false
		}
		must(err)
	}
	return dbProcessToStorage(dbProcess), true
}

func uuidSliceToInterfaceSlice(inlist []api.UUID) (outlist []interface{}) {
	for _, v := range inlist {
		outlist = append(outlist, uuid.UUID(v))
	}
	return
}

func processStatusSliceToInterfaceSlice(inlist []api.Process_Status) (outlist []interface{}) {
	for _, v := range inlist {
		outlist = append(outlist, v)
	}
	return
}

// Query lookup for processes
func (s *ProcessStorage) Query(filter api.ProcessFilter) []storage.Process {
	var dbProcessList []model.Process
	tx := s.MustBegin()
	defer tx.Rollback()
	q := tx.Query(s.model.Process).
		OrderBy(s.model.Process.CreatedAt)
	if len(filter.ID) != 0 {
		q = q.Filter(s.model.Process.ID.In(
			uuidSliceToInterfaceSlice(filter.ID)...))
	}
	if len(filter.TriggerEmitterID) != 0 {
		q = q.Filter(s.model.Process.TriggerEmitterID.In(
			uuidSliceToInterfaceSlice(filter.TriggerEmitterID)...))
	}
	if len(filter.TriggerEnvelopeID) != 0 {
		q = q.Filter(s.model.Process.TriggerEnvelopeID.In(
			uuidSliceToInterfaceSlice(filter.TriggerEnvelopeID)...))
	}
	if len(filter.PipelineID) != 0 {
		q = q.Filter(s.model.Process.PipelineID.In(
			uuidSliceToInterfaceSlice(filter.PipelineID)...))
	}
	if filter.CreatedFrom != nil {
		q = q.Filter(s.model.Process.CreatedAt.Gte(filter.CreatedFrom))
	}
	if filter.CreatedTo != nil {
		q = q.Filter(s.model.Process.CreatedAt.Lte(filter.CreatedTo))
	}

	var joinedState bool

	if len(filter.Status) != 0 ||
		(filter.StatusChangedFrom != nil) ||
		(filter.StatusChangedTo != nil) {

		q = q.LeftJoin(s.model.ProcessState)
		joinedState = true

		if len(filter.Status) != 0 {
			q = q.Filter(
				s.model.ProcessState.Status.In(
					processStatusSliceToInterfaceSlice(filter.Status)...),
			)
		}

		if filter.StatusChangedFrom != nil {
			q = q.Filter(
				s.model.ProcessState.StatusChanged.Gte(filter.StatusChangedFrom),
			)
		}
		if filter.StatusChangedTo != nil {
			q = q.Filter(
				s.model.ProcessState.StatusChanged.Lte(filter.StatusChangedTo),
			)
		}
	}

	if len(filter.ResultAcked) != 0 {
		var (
			includeTrue  bool
			includeFalse bool
		)
		for _, v := range filter.ResultAcked {
			if v {
				includeTrue = true
			} else {
				includeFalse = true
			}
		}
		if includeTrue != includeFalse {
			if !joinedState {
				q = q.LeftJoin(s.model.ProcessState)
				joinedState = true
			}
			q = q.Filter(
				s.model.ProcessState.ResultAcked.Eq(includeTrue),
			)
		}
	}

	if len(filter.EnvelopeID) != 0 {

		q = q.Filter(qb.Or(
			qb.Exists(
				s.model.Route.GetMapper().Table().Select(qb.SQLText("1")).
					Where(s.model.Route.ProcessID.Eq(s.model.Process.ID),
						s.model.Route.EnvelopeID.In(
							uuidSliceToInterfaceSlice(filter.EnvelopeID)...),
					),
			),
			s.model.Process.TriggerEmitterID.In(
				uuidSliceToInterfaceSlice(filter.EnvelopeID)...),
		))
	}

	must(
		q.All(&dbProcessList))
	var pList []storage.Process
	for _, dbP := range dbProcessList {
		pList = append(pList, dbProcessToStorage(dbP))
	}
	return pList
}

// FindProcessForEnvelope finds the process(es) that reference a bunch of envelopes
func (s *ProcessStorage) FindProcessForEnvelope(
	envelopeIDs []api.UUID,
) map[api.UUID]api.UUID {
	result := make(map[api.UUID]api.UUID)
	tx := s.MustBegin()
	defer tx.Rollback()

	q := tx.Query(s.model.Route).
		Select(s.model.Route.EnvelopeID, s.model.Route.ProcessID).
		Where(s.model.Route.EnvelopeID.In(uuidSliceToInterfaceSlice(envelopeIDs)...))

	rows, err := q.SQLQuery()
	must(err)
	defer rows.Close()
	for rows.Next() {
		var eID, pID uuid.UUID
		must(rows.Scan(&eID, &pID))
		result[api.UUID(eID)] = api.UUID(pID)
	}

	// Some process may be route less, we need their trigger envelope ids
	q = tx.Query(s.model.Process).
		Select(s.model.Process.TriggerEnvelopeID, s.model.Process.ID).
		Where(s.model.Process.TriggerEnvelopeID.In(uuidSliceToInterfaceSlice(envelopeIDs)...))

	rows, err = q.SQLQuery()
	must(err)
	defer rows.Close()
	for rows.Next() {
		var eID, pID uuid.UUID
		must(rows.Scan(&eID, &pID))
		result[api.UUID(eID)] = api.UUID(pID)
	}

	return result
}

// CreateProcess saves a process
func (s *ProcessStorage) CreateProcess(p storage.Process) (api.UUID, error) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	p.ID = api.UUID{}

	dbProcess := storageProcessToDB(p)

	must(tx.Insert(&dbProcess))

	tx.MustCommit()

	return api.UUID(dbProcess.ID), nil
}

// SetPipeline updates the pipeline & sourceOutput of a process
func (s *ProcessStorage) SetPipeline(
	processID api.UUID,
	pipelineID api.UUID,
	sourceOutput storage.NodeOutputRef,
) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	p, ok := s.loadProcess(tx.Tx, processID)
	if !ok {
		panic("Unknown processID")
	}
	p.PipelineID = pipelineID
	p.SourceOutput = sourceOutput
	dbP := storageProcessToDB(p)
	must(tx.Update(
		&dbP,
		model.ProcessPipelineID,
		model.ProcessSourceNodeID,
		model.ProcessSourceSlotID))
	tx.MustCommit()
}

// GetTargets returns the targets for an envelope if already known, or 'nil' if not
func (s *ProcessStorage) GetTargets(
	processID api.UUID,
	source storage.NodeOutputRef,
	envelopeID api.UUID,
	noRouteTableUpdate bool,
	calcTargets func() ([]api.EnvelopeTarget, bool, error),
) (targets []api.EnvelopeTarget, calcErr error) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	bSource, err := json.Marshal(source)
	if err != nil {
		panic(err)
	}

	Route, RouteTarget := s.model.Route, s.model.RouteTarget

	var dbTargetList []model.RouteTarget

	if err := tx.Query(RouteTarget).
		LeftJoin(Route).
		Where(
			Route.ProcessID.Eq(uuid.UUID(processID)),
			Route.Source.Eq(bSource),
			Route.EnvelopeID.Eq(uuid.UUID(envelopeID))).
		All(&dbTargetList); err != nil {
		panic(err)
	}

	if len(dbTargetList) != 0 {
		for _, dbTarget := range dbTargetList {
			targets = append(targets, api.EnvelopeTarget{
				ActorID: api.UUID(dbTarget.ActorID),
				NodeID:  dbTarget.NodeID,
				Input:   dbTarget.Input,
			})
		}
		return
	}

	tx.Rollback()

	var isResponse bool
	targets, isResponse, calcErr = calcTargets()
	if calcErr != nil || noRouteTableUpdate {
		return
	}
	route := model.Route{
		ProcessID:  uuid.UUID(processID),
		Source:     string(bSource),
		EnvelopeID: uuid.UUID(envelopeID),
	}

	if isResponse {
		tx = s.MustBegin()
		// Save the response
		if _, err := tx.GetEngine().Exec(model.RawSQL(
			"UPDATE process_state SET response = $1",
			uuid.UUID(envelopeID),
		)); err != nil {
			tx.Rollback()
			panic(err)
		}
		tx.MustCommit()
	}

	tx2 := s.MustBegin()
	defer tx2.RollbackIfOpened()

	// Use an exclusive transaction
	// on PG: SET TRANSACTION SERIALIZABLE
	if false && s.db.Engine.Dialect().Driver() == "postgres" {
		_, err := tx2.GetEngine().Exec(model.RawSQL("SET TRANSACTION ISOLATION LEVEL SERIALIZABLE"))
		if err != nil {
			panic(err)
		}
	}

	if err := tx2.Query(RouteTarget).
		LeftJoin(Route).
		Where(
			Route.ProcessID.Eq(uuid.UUID(processID)),
			Route.Source.Eq(bSource),
			Route.EnvelopeID.Eq(uuid.UUID(envelopeID))).
		All(&dbTargetList); err != nil {
		panic(err)
	}

	// The routes where calculated faster by someone else
	// We return their result and ignore ours
	if len(dbTargetList) != 0 {
		for _, dbTarget := range dbTargetList {
			targets = append(targets, api.EnvelopeTarget{
				ActorID: api.UUID(dbTarget.ActorID),
				NodeID:  dbTarget.NodeID,
				Input:   dbTarget.Input,
			})
		}
		return
	}

	var logentries []storage.ProcessLogEntry
	// TODO: detect a concurrent insert and handle it gently
	must(tx2.Insert(&route))
	for _, target := range targets {
		must(tx2.Insert(&model.RouteTarget{
			RouteID: route.ID,
			ActorID: uuid.UUID(target.ActorID),
			NodeID:  target.NodeID,
			Input:   target.Input,
		}))
		logentry, err := s.processLogEmission(
			tx2,
			processID,
			storage.Emission{
				ProcessID:   processID,
				EnvelopeID:  envelopeID,
				TargetActor: target.ActorID,
				TargetInputRef: storage.NodeInputRef{
					NodeID: target.NodeID,
					Input:  target.Input,
				},
			},
			storage.EnvelopeStartPos())
		must(err)
		logentries = append(logentries, logentry)
	}
	s.logger.Noticef("New routes -> logentries = %+v", logentries)
	must(s.updateStateAndCommit(tx2, processID, logentries))
	return

}

// UpdateReceiveStatus updates a reception status
func (s *ProcessStorage) UpdateReceiveStatus(reception storage.Reception, status storage.ReceptionStatus) error {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	logentry, err := s.processLogReception(tx, reception.ProcessID, reception, status)
	must(err)

	return s.updateStateAndCommit(
		tx, reception.ProcessID, []storage.ProcessLogEntry{logentry})
}

// ProcessNodeEnd signals the end of a ProcessNode.
func (s *ProcessStorage) ProcessNodeEnd(
	processID api.UUID, node storage.ProcessNode,
	status api.ActorProcessingState_Status,
) error {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	logentry, err := s.processLogProcessNodeEnd(
		tx, processID,
		storage.ProcessNodeEnd{
			Node:   node,
			Status: status,
		})
	must(err)

	return s.updateStateAndCommit(
		tx, processID, []storage.ProcessLogEntry{logentry})
}

// GetInputs returns the incoming routes for a processnode
func (s *ProcessStorage) GetInputs(
	processID api.UUID, targetNode storage.ProcessNode,
) (inputs []storage.ProcessNodeInput) {
	tx := s.MustBegin()
	defer tx.Rollback()

	q := tx.Query(s.model.RouteTarget).
		LeftJoin(s.model.Route).
		Select(s.model.RouteTarget.Input, s.model.Route.EnvelopeID).
		Where(
			s.model.Route.ProcessID.Eq(uuid.UUID(processID)),
			s.model.RouteTarget.ActorID.Eq(uuid.UUID(targetNode.ActorID)),
			s.model.RouteTarget.NodeID.Eq(targetNode.NodeID),
		)
	rows, err := q.SQLQuery()
	must(err)
	defer rows.Close()

	for rows.Next() {
		var input storage.ProcessNodeInput
		must(rows.Scan(&input.Name, (*uuid.UUID)(&input.EnvelopeID)))
		inputs = append(inputs, input)
	}

	return
}

// GetState returns an instant copy of the state of a process
func (s *ProcessStorage) GetState(processID api.UUID) storage.ProcessState {
	tx := s.MustBegin()
	defer tx.Rollback()
	state, err := s.loadState(tx.Tx, processID, false)
	must(err)
	return state
}

// GetStateUpdater returns a StateUpdater for a process
func (s *ProcessStorage) GetStateUpdater(processID api.UUID) storage.StateUpdater {
	tx := s.MustBegin()
	updater, err := NewStateUpdater(s, tx, processID)
	if err != nil {
		panic(err)
	}
	return updater
}

func isValidTransition(from, to api.Process_Status) bool {
	switch from {
	case api.Process_NOSTATUS,
		api.Process_INITIAL,
		api.Process_RUNNING:
		return to > from
	case api.Process_PAUSED:
		return to >= api.Process_RUNNING
	case api.Process_DONE:
		return to == from
	case api.Process_ERROR:
		return to == from
	}
	return false
}

// SetStatus changes the current status. Only Runnning->Paused and Paused->Running
// are allowed at this point
func (s *ProcessStorage) SetStatus(
	processID api.UUID, status api.Process_Status, reason string,
) error {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	process, ok := s.loadProcess(tx.Tx, processID)
	if !ok {
		return fmt.Errorf("No such process: %s", processID)
	}

	state, err := s.loadState(tx.Tx, processID, false)
	must(err)

	if !isValidTransition(state.Status, status) {
		return fmt.Errorf("Invalid state transition: %s -> %s", state.Status, status)
	}

	old := state.Status
	state.Status = status
	state.StatusReason = reason

	err = s.saveState(tx.Tx, processID, state)
	must(err)

	tx.MustCommit()

	s.onStatusChanged(process, old, state.Status)

	return nil
}

func (s *ProcessStorage) loadState(tx *yago.Tx, processID api.UUID, forUpdate bool) (pState storage.ProcessState, err error) {
	var dbProcessState model.ProcessState

	q := tx.Query(s.model.ProcessState)
	if forUpdate {
		q = q.ForUpdate()
	}
	if err = q.Get(&dbProcessState, uuid.UUID(processID)); err != nil && err != yago.ErrRecordNotFound {
		return
	}

	if err == yago.ErrRecordNotFound {
		pState.Status = api.Process_INITIAL
	} else {
		pState.Status = dbProcessState.Status
		pState.StatusChanged = dbProcessState.StatusChanged
		pState.StatusReason = dbProcessState.StatusReason
		pState.ResultAcked = dbProcessState.ResultAcked
		if err = json.Unmarshal(
			[]byte(dbProcessState.NodeStates),
			&pState.NodeStates); err != nil {
			return
		}
		pState.NodeStatesChanged = dbProcessState.NodeStatesChanged
		if dbProcessState.Response.Valid {
			pState.ResponseEnvelopeID = api.UUID(dbProcessState.Response.UUID)
		}
	}
	return pState, nil
}

func (s *ProcessStorage) saveState(tx *yago.Tx, processID api.UUID, pState storage.ProcessState) error {
	b, err := json.Marshal(pState.NodeStates)
	if err != nil {
		return err
	}
	dbProcessState := model.ProcessState{
		ProcessID:         uuid.UUID(processID),
		Status:            pState.Status,
		StatusChanged:     pState.StatusChanged,
		StatusReason:      pState.StatusReason,
		NodeStates:        string(b),
		NodeStatesChanged: pState.NodeStatesChanged,
		ResultAcked:       pState.ResultAcked,
	}
	if !pState.ResponseEnvelopeID.IsZero() {
		dbProcessState.Response.UUID = uuid.UUID(pState.ResponseEnvelopeID)
		dbProcessState.Response.Valid = true
	}
	exists, err := tx.Query(s.model.ProcessState).
		Where(s.model.ProcessState.ProcessID.Eq(uuid.UUID(processID))).
		Exists()
	if err != nil {
		return err
	}
	if !exists {
		return tx.Insert(&dbProcessState)
	}
	return tx.Update(&dbProcessState)
}

func (s *ProcessStorage) processLogEmission(tx *TxHelper, processID api.UUID, emission storage.Emission, pos storage.EnvelopeReadPos) (storage.ProcessLogEntry, error) {
	logentry := storage.ProcessLogEntry{
		ProcessID:        processID,
		Emission:         &emission,
		EmissionPosition: &pos,
	}
	return s.processLog(tx, processID, logentry)
}

func (s *ProcessStorage) processLogReception(tx *TxHelper, processID api.UUID, reception storage.Reception, status storage.ReceptionStatus) (storage.ProcessLogEntry, error) {
	logentry := storage.ProcessLogEntry{
		ProcessID:       processID,
		Reception:       &reception,
		ReceptionStatus: status,
	}
	return s.processLog(tx, processID, logentry)
}

func (s *ProcessStorage) processLogProcessNodeEnd(
	tx *TxHelper, processID api.UUID, nodeEnd storage.ProcessNodeEnd,
) (storage.ProcessLogEntry, error) {
	return s.processLog(
		tx, processID,
		storage.ProcessLogEntry{
			ProcessID:      processID,
			ProcessNodeEnd: &nodeEnd,
		})
}

func (s *ProcessStorage) processLog(tx *TxHelper, processID api.UUID, logentry storage.ProcessLogEntry) (storage.ProcessLogEntry, error) {
	logentry.ProcessID = processID
	logentry.Timestamp = time.Now()
	b, err := json.Marshal(logentry)
	if err != nil {
		return storage.ProcessLogEntry{}, err
	}
	dbLogEntry := model.ProcessLogEntry{
		ProcessID: uuid.UUID(processID),
		Timestamp: logentry.Timestamp,
		Data:      string(b),
	}
	if err := tx.Insert(&dbLogEntry); err != nil {
		return storage.ProcessLogEntry{}, err
	}
	logentry.ID = strconv.FormatInt(dbLogEntry.ID, 10)
	return logentry, nil
}

// AckResult ...
func (s *ProcessStorage) AckResult(processID api.UUID) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	logentry, err := s.processLog(tx, processID, storage.ProcessLogEntry{
		ResultAck: true,
	})
	must(err)
	tx.MustCommit()

	s.processLogEntryHandler([]storage.ProcessLogEntry{logentry})
}

func (s *ProcessStorage) updateStateAndCommit(tx *TxHelper, processID api.UUID, logentries []storage.ProcessLogEntry) error {
	// Commit the transaction so nextState is called outside it
	tx.MustCommit()

	s.processLogEntryHandler(logentries)
	return nil
}

// GetUnprocessedProcessLogEntries returns the unprocessed log entries
func (s *ProcessStorage) GetUnprocessedProcessLogEntries() (entries []storage.ProcessLogEntry) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	var dbLogEntries []model.ProcessLogEntry

	q := tx.Query(s.model.ProcessLogEntry).
		Where(s.model.ProcessLogEntry.LogProcessed.Eq(false))
	must(q.All(&dbLogEntries))

	must(tx.Rollback())

	for _, dbLogEntry := range dbLogEntries {
		var logEntry storage.ProcessLogEntry
		if err := json.Unmarshal([]byte(dbLogEntry.Data), &logEntry); err != nil {
			panic(err)
		}
		logEntry.ID = strconv.FormatInt(dbLogEntry.ID, 10)
		logEntry.ProcessID = api.UUID(dbLogEntry.ProcessID)
		entries = append(entries, logEntry)
	}

	return
}

// GetProcessLogEntries returns the unprocessed log entries
func (s *ProcessStorage) GetProcessLogEntries(id api.UUID) (entries []storage.ProcessLogEntry) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	var dbLogEntries []model.ProcessLogEntry

	q := tx.Query(s.model.ProcessLogEntry).
		Where(s.model.ProcessLogEntry.ProcessID.Eq(uuid.UUID(id)))
	must(q.All(&dbLogEntries))

	must(tx.Rollback())

	for _, dbLogEntry := range dbLogEntries {
		var logEntry storage.ProcessLogEntry
		if err := json.Unmarshal([]byte(dbLogEntry.Data), &logEntry); err != nil {
			panic(err)
		}
		logEntry.ID = strconv.FormatInt(dbLogEntry.ID, 10)
		logEntry.ProcessID = api.UUID(dbLogEntry.ProcessID)
		entries = append(entries, logEntry)
	}

	return
}

// QueryPostMortem returns a list of dead processes
func (s *ProcessStorage) QueryPostMortem(level api.LogLevel, statusList ...api.PMProcess_Status) []storage.Process {
	tx := s.MustBegin()
	defer tx.Rollback()

	q := tx.Query(s.model.Process).
		LeftJoin(s.model.ProcessState)
	if len(statusList) != 0 {
		var crits []qb.Clause
		for _, status := range statusList {
			crits = append(crits, s.model.ProcessState.PMStatus.Eq(status))
		}
		q = q.Filter(qb.Or(crits...))
	}
	if level == api.LogLevel_ERROR {
		q = q.Filter(s.model.ProcessState.PMLevel.Eq(api.LogLevel_ERROR))
	}

	var dbProcessList []model.Process
	must(
		q.All(&dbProcessList))
	var pList []storage.Process
	for _, dbP := range dbProcessList {
		pList = append(pList, dbProcessToStorage(dbP))
	}
	return pList
}

// GetPostMortemState returns the postmortem state of a process
func (s *ProcessStorage) GetPostMortemState(id api.UUID) storage.PostMortemState {
	tx := s.MustBegin()
	defer tx.Rollback()

	row := tx.Query(s.model.ProcessState).
		Where(s.model.ProcessState.ProcessID.Eq(uuid.UUID(id))).
		Select(
			s.model.ProcessState.PMLevel,
			s.model.ProcessState.PMStatus,
			s.model.ProcessState.PMComment,
			s.model.ProcessState.PMChanged).
		SQLQueryRow()

	var pmLevel, pmStatus []byte
	var pmState storage.PostMortemState

	must(row.Scan(&pmLevel, &pmStatus, &pmState.Comment, &pmState.Changed))
	must(pmState.Level.UnmarshalText(pmLevel))
	must(pmState.Status.UnmarshalText(pmStatus))

	return pmState
}

// SetPostMortemState sets the postmortem state of a process
func (s *ProcessStorage) SetPostMortemState(id api.UUID, pmState storage.PostMortemState) {
	tx := s.MustBegin()
	defer tx.MustCommit()

	pState := model.ProcessState{
		ProcessID: uuid.UUID(id),
		PMLevel:   pmState.Level,
		PMStatus:  pmState.Status,
		PMComment: pmState.Comment,
		PMChanged: pmState.Changed,
	}
	must(tx.Update(&pState,
		model.ProcessStatePMLevel,
		model.ProcessStatePMStatus,
		model.ProcessStatePMComment,
		model.ProcessStatePMChanged,
	))
}

// PurgeProcess removes all data of a process.
// The process must be terminated
// TODO Make a version of the function that takes several ids, and return
// the list of actually purged processes.
func (s *ProcessStorage) PurgeProcess(ids []api.UUID) ([]api.UUID, error) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	var purged = make([]api.UUID, len(ids))
	var dbIDs = make([]uuid.UUID, len(ids))
	for i, id := range ids {
		dbIDs[i] = uuid.UUID(id)
	}

	rows, err := tx.GetEngine().Query(model.RawSQL(
		"DELETE FROM process "+
			"WHERE id = ANY($1) "+
			"AND EXISTS("+
			" SELECT 1 "+
			" FROM process_state "+
			" WHERE process_state.process_id = process.id "+
			"   AND process_state.status IN ('DONE', 'ERROR')"+
			") "+
			"RETURNING (id)",
		pq.Array(dbIDs),
	),
	)
	if err != nil {
		return nil, err
	}
	defer rows.Close()

	var count = 0
	for ; rows.Next(); count++ {
		if err := rows.Scan((*uuid.UUID)(&purged[count])); err != nil {
			return nil, err
		}
	}

	purged = purged[:count]

	tx.MustCommit()

	return purged, nil
}

// StateUpdater implements storage.StateUpdater
type StateUpdater struct {
	processStorage *ProcessStorage
	tx             *TxHelper
	processID      api.UUID
	state          storage.ProcessState

	pos        storage.EnvelopeReadPos
	logentries []storage.ProcessLogEntry
	lastError  error
}

// NewStateUpdater creates a new StateUpdater
func NewStateUpdater(
	processStorage *ProcessStorage, tx *TxHelper, processID api.UUID) (*StateUpdater, error) {
	state, err := processStorage.loadState(tx.Tx, processID, true)
	if err != nil {
		return nil, err
	}
	return &StateUpdater{
		processStorage: processStorage,
		tx:             tx,
		processID:      processID,
		state:          state,
	}, nil
}

// Get returns the current state
func (u *StateUpdater) Get() (storage.ProcessState, error) {
	return u.state, nil
}

// Set updates the state
func (u *StateUpdater) Set(state storage.ProcessState) error {
	err := u.processStorage.saveState(u.tx.Tx, u.processID, state)
	if err != nil {
		return err
	}
	u.state = state
	return nil
}

// MarkLogEntry marks a ProcessLogEntry as processed
func (u *StateUpdater) MarkLogEntry(id string, withError bool) error {
	dbID, err := strconv.ParseInt(id, 10, 64)
	if err != nil {
		return err
	}
	return u.tx.Update(
		&model.ProcessLogEntry{ID: dbID, LogProcessed: true, ProcessingError: withError},
		model.ProcessLogEntryLogProcessed,
		model.ProcessLogEntryProcessingError,
	)
}

// Unlock release the lock on the data
func (u *StateUpdater) Unlock() error {
	return u.tx.Commit()
}
