// Copyright 2019 Orus.io. All rights reserved.

package sql

import (
	"testing"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
	"bitbucket.org/orus-io/xbus/testutils"
	"bitbucket.org/orus-io/xbus/testutils/sqldb"
	storagetest "bitbucket.org/orus-io/xbus/testutils/storage"
	"github.com/m4rw3r/uuid"
)

func makeEnvelopeStorageFactory(tb testing.TB) (
	func(tb testing.TB) (storage.ProcessStorage, func()),
	func(),
) {
	testDB := sqldb.GetTestDB(tb)
	if testDB == nil {
		tb.Fatal("Could not init testdb")
	}
	m, db := testDB.GetModelDB(1)

	// TODO There constraints should probably be removed from the database
	// Meanwhile, they are in the way of keeping process storage tests simples
	if _, err := db.Engine.DB().Exec(
		"ALTER TABLE process DROP CONSTRAINT process_trigger_envelope_id_fkey",
	); err != nil {
		tb.Fatal(err)
	}
	if _, err := db.Engine.DB().Exec(
		"ALTER TABLE process DROP CONSTRAINT process_trigger_emitter_id_fkey",
	); err != nil {
		tb.Fatal(err)
	}

	return func(tb testing.TB) (storage.ProcessStorage, func()) {
			return NewProcessStorage(m, db, testutils.GetTestLogger(tb)), func() {
				if _, err := db.Engine.DB().Exec("DELETE FROM process_state"); err != nil {
					tb.Fatal(err)
				}
				if _, err := db.Engine.DB().Exec("DELETE FROM process"); err != nil {
					tb.Fatal(err)
				}
			}
		}, func() {
			db.Close()
			testDB.Close()
		}
}

func TestProcessStorage(t *testing.T) {
	factory, clean := makeEnvelopeStorageFactory(t)
	defer clean()

	storagetest.ProcessStorageTests(t, factory)
}

func massCreateProcess(ps storage.ProcessStorage, count int, status api.Process_Status) []api.UUID {
	sqlPs := ps.(*ProcessStorage)
	tx := sqlPs.MustBegin()
	defer tx.RollbackIfOpened()
	rows, err := tx.GetEngine().Query(model.RawSQL(`
WITH s AS (SELECT generate_series(1,$1))
INSERT INTO process (
    id,
    created_at,
    trigger_emitter_id,
    trigger_envelope_id,
    pipeline_id,
    source_node_id,
    source_slot_id
) SELECT
    gen_random_uuid(),
    now(),
    '00000000-0000-0000-0000-000000000000'::uuid,
    '00000000-0000-0000-0000-000000000000'::uuid,
    '00000000-0000-0000-0000-000000000000'::uuid,
    '00000000-0000-0000-0000-000000000000'::uuid,
    '00000000-0000-0000-0000-000000000000'::uuid
    FROM s

RETURNING (id);
`, count))
	if err != nil {
		panic(err)
	}
	var ids = make([]api.UUID, count)
	for i := 0; rows.Next(); i++ {
		if err := rows.Scan((*uuid.UUID)(&ids[i])); err != nil {
			panic(err)
		}
	}

	_, err = tx.GetEngine().Exec(model.RawSQL(
		`WITH p AS (
    SELECT id
    FROM process
    LEFT JOIN process_state ON process.id = process_state.process_id
    WHERE process_state.process_id IS NULL
)
INSERT INTO process_state (
    process_id,
    status,
    status_reason,
    node_states_changed,
    pm_level,
    pm_comment,
    pm_changed,
    status_changed,
    node_states,
    pm_status
) SELECT
    id,
    'DONE',
    '',
    'EPOCH',
    '',
    '',
    'EPOCH',
    now(),
    '',
    ''
FROM p
`))
	if err != nil {
		panic(err)
	}
	tx.Commit()

	return ids
}

func BenchmarkProcessStorage(b *testing.B) {
	factory, clean := makeEnvelopeStorageFactory(b)
	defer clean()

	storagetest.ProcessStorageBenchmark(b, factory, massCreateProcess)
}
