// Copyright 2016 Orus.io. All rights reserved.

package mocks

import (
	"fmt"
	"math"
	"sync"
	"time"

	xbus "bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
)

// NewEnvelopeStorage init an empty EnvelopeStorage suitable for tests
func NewEnvelopeStorage(logger xbus.Logger) *EnvelopeStorage {
	return &EnvelopeStorage{
		receivers:   make(map[api.UUID]*xbus.EnvelopeReceiver),
		envelopes:   make(map[api.UUID]api.Envelope),
		lastReceive: make(map[api.UUID]time.Time),
		isStalled:   make(map[api.UUID]bool),
		logger:      logger,
	}
}

// EnvelopeStorage implements EnvelopeStorage on top of a SQL database
type EnvelopeStorage struct {
	receivers   map[api.UUID]*xbus.EnvelopeReceiver
	envelopes   map[api.UUID]api.Envelope
	lastReceive map[api.UUID]time.Time
	isStalled   map[api.UUID]bool
	logger      xbus.Logger
	evHandler   storage.EnvelopeStorageEventHandler

	m sync.Mutex
}

// SetEventHandler sets the event handler
func (s *EnvelopeStorage) SetEventHandler(handler storage.EnvelopeStorageEventHandler) {
	s.evHandler = handler
}

// GetEventTypes returns an envelop event types
func (s *EnvelopeStorage) GetEventTypes(id api.UUID) ([]string, bool) {
	r, ok := s.receivers[id]
	if !ok {
		return nil, false
	}
	eventtypes := r.EventTypes()
	return eventtypes, len(eventtypes) != 0
}

func (s *EnvelopeStorage) receiveEnvelope(id api.UUID) {
	envelope, ok := s.envelopes[id]
	r := s.receivers[id]

	if !ok {
		envelope.ID = r.ID
	}

	for _, event := range r.PartialEvents() {
		var index = -1
		for i, tgt := range envelope.Events {
			if tgt.ID == event.ID {
				index = i
				break
			}
		}
		if index == -1 {
			index = len(envelope.Events)
			envelope.Events = append(envelope.Events, api.Event{})
		}

		envelope.Events[index].ID = event.ID
		envelope.Events[index].Type = event.Type
		envelope.Events[index].ItemCount = event.ItemCount
		envelope.Events[index].Checksum = event.Checksum
		envelope.Events[index].Items = append(
			envelope.Events[index].Items,
			r.NextItems(event.Type, int(math.MaxInt64))...)
	}

	s.envelopes[id] = envelope
}

// StoreEnvelope store a part of an envelope and returns its status
// in the storage
func (s *EnvelopeStorage) StoreEnvelope(e api.Envelope) (storage.EnvelopeState, error) {
	s.m.Lock()
	defer s.m.Unlock()

	var status = storage.EnvelopeState{ID: e.ID}

	r, ok := s.receivers[e.ID]
	if !ok {
		newr, err := xbus.NewEnvelopeReceiver("", e)
		if err != nil {
			status.Status = storage.EnvelopeStorageReceptionError
			status.Err = err
			return status, err
		}
		s.receivers[e.ID] = newr
		r = newr
	} else {
		if err := r.AddFragment(e); err != nil {
			status.Status = storage.EnvelopeStorageReceptionError
			status.Err = err
			return status, err
		}
	}

	s.isStalled[e.ID] = false
	s.lastReceive[e.ID] = time.Now()

	status.TypesKnown = r.EventTypes() != nil
	switch r.Status() {
	case api.EnvelopeAck_RECEIVING:
		status.Status = storage.EnvelopeStorageReceptionReceiving
	case api.EnvelopeAck_ACCEPTED:
		status.Status = storage.EnvelopeStorageReceptionComplete
	case api.EnvelopeAck_ERROR:
		status.Status = storage.EnvelopeStorageReceptionError
	}

	s.receiveEnvelope(e.ID)

	envelope := s.envelopes[e.ID]

	var newData bool

	for _, ev := range envelope.Events {
		if len(ev.Items) != 0 {
			newData = true
		}
		evS := storage.EventState{
			ID:             ev.ID,
			Type:           ev.Type,
			AvailableItems: uint64(len(ev.Items)),
			TotalItems:     ev.ItemCount,
		}
		if evS.TotalItems != 0 && evS.TotalItems == evS.AvailableItems {
			evS.Status = storage.EnvelopeStorageReceptionComplete
		} else {
			evS.Status = storage.EnvelopeStorageReceptionReceiving
		}
		status.Events = append(status.Events, evS)
	}

	if s.evHandler != nil {
		s.evHandler(status.ID, status.Status, newData)
	}

	return status, nil
}

// CheckStalled ...
func (s *EnvelopeStorage) CheckStalled(id api.UUID, delay time.Duration) bool {
	s.m.Lock()
	defer s.m.Unlock()

	_, ok := s.envelopes[id]
	if !ok {
		return false
	}
	if stalled, _ := s.isStalled[id]; stalled {
		return true
	}
	lastReceive := s.lastReceive[id]
	if time.Now().Sub(lastReceive) > delay {
		s.isStalled[id] = true
		s.evHandler(id, storage.EnvelopeStorageReceptionStalled, false)
		return true
	}
	return false
}

// ReadEnvelope reads as much as it can of an envelope, starting at 'position'
// The returned envelope should be suitable for encrypted nats transfert, ie
// its total jsonified size should be under maxsize
func (s *EnvelopeStorage) ReadEnvelope(id api.UUID, position storage.EnvelopeReadPos, maxsize int) (api.Envelope, storage.EnvelopeReadPos, error) {
	s.m.Lock()
	defer s.m.Unlock()

	envelope, ok := s.envelopes[id]
	if !ok {
		return api.Envelope{}, position, fmt.Errorf("No such envelope: %s", id)
	}

	if position.Start {
		for _, e := range envelope.Events {
			position.Events[e.ID] = storage.EventReadPos{}
		}
		position.Start = false
	}
	fragment := api.Envelope{
		ID: envelope.ID,
	}
	for _, event := range envelope.Events {
		fragment.EventIDs = append(fragment.EventIDs, event.ID)
	}

	// TODO in case we already know the envelope is small enough, return
	// it complete at once

	// approx size of the jsonified envelope before any event is added
	var estimateSize = 36 + len(fragment.EventIDs)*3

	position = position.Clone()

	tgtEvents := map[api.UUID]api.Event{}

	for _, ev := range envelope.Events {
		eventPos := position.Events[ev.ID]
		if eventPos.ItemCount < uint64(len(ev.Items)) {
			eventPos.Index++
			tgtEvents[ev.ID] = api.Event{
				ID:        ev.ID,
				Type:      ev.Type,
				Index:     eventPos.Index,
				ItemCount: ev.ItemCount,
				Checksum:  ev.Checksum,
			}
			estimateSize += 36 + len(ev.Type) + 8
		}
		position.Events[ev.ID] = eventPos
	}

	for maxsize == 0 || estimateSize < maxsize {
		before := estimateSize
		for _, ev := range envelope.Events {
			pos := position.Events[ev.ID]
			if pos.ItemCount == uint64(len(ev.Items)) {
				continue
			}
			tgt := tgtEvents[ev.ID]
			item := ev.Items[pos.ItemCount]
			tgt.Items = append(tgt.Items, item)
			pos.ItemCount++

			estimateSize += int(float64(len(item)) * 1.1)

			position.Events[ev.ID] = pos
			tgtEvents[ev.ID] = tgt
			if maxsize != 0 && estimateSize >= maxsize {
				break
			}
		}
		if estimateSize == before {
			break
		}
	}

	for _, ev := range tgtEvents {
		fragment.Events = append(fragment.Events, ev)
	}

	complete := true
	for _, ev := range envelope.Events {
		eventPos := position.Events[ev.ID]
		if ev.ItemCount == 0 || eventPos.ItemCount != ev.ItemCount {
			complete = false
		}
	}

	if complete {
		fragment.Last = true
		position.Complete = true
	}

	return fragment, position, nil
}

// Purge removes the envelopes fitered by the given filter
func (s *EnvelopeStorage) Purge(filter storage.EnvelopeFilter) {
	var todelete []api.UUID

	for id := range s.envelopes {
		todelete = append(todelete, filter([]api.UUID{id})...)
	}

	for _, id := range todelete {
		delete(s.envelopes, id)
		delete(s.receivers, id)
	}
}
