// Copyright 2016 Orus.io. All rights reserved.

package sqldb

import (
	"os"
	"testing"

	// load postgres driver
	_ "github.com/lib/pq"

	// load sqlite driver
	_ "github.com/mattn/go-sqlite3"
	"github.com/orus-io/yago"
	"github.com/slicebit/qb"

	xbus "bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/xbus/migration"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
	"bitbucket.org/orus-io/xbus/testutils"
)

const testMaxDBConnections = 10

// CloseFunction is a function with no arg nor return values for closing a resource
type CloseFunction func()

type migrateLogger struct {
	log xbus.Logger
}

func (l migrateLogger) Printf(format string, v ...interface{}) {
	l.log.Noticef(format, v...)
}

func (migrateLogger) Verbose() bool {
	return false
}

// TestDB holds a test database and provides access to it until its 'Close'
// function is called
type TestDB struct {
	Driver   string
	Dsn      string
	LogFlags qb.LogFlags
	Logger   xbus.Logger

	t    testing.TB
	lock *testutils.PortLock
}

// Close releases the underlying database
func (tdb *TestDB) Close() {
	if tdb.lock != nil {
		if err := tdb.lock.Unlock(); err != nil {
			tdb.t.Errorf("Error releasing lock: %s", err)
		}
	}
}

// GetModelDB returns a new connection pool to the database
func (tdb *TestDB) GetModelDB(maxConn int) (*model.Model, *yago.DB) {
	m, db, err := model.NewDB(tdb.Driver, tdb.Dsn, maxConn, tdb.Logger, tdb.LogFlags)
	if err != nil {
		tdb.t.Fatalf("Cannot initialize test database: %s", err)
		return nil, nil
	}
	return m, db
}

// GetTestDB creates a test database and returns a TestDB pointing to it
func GetTestDB(t testing.TB) *TestDB {
	var (
		driver     = "sqlite3"
		droptables = false
	)

	dsn, ok := os.LookupEnv("XBUS_TEST_POSTGRES_DSN")
	if ok {
		driver = "postgres"
		droptables = true
	} else {
		dsn, ok = os.LookupEnv("XBUS_TEST_SQLITE_DSN")
		if !ok {
			os.Remove("test.db")
			dsn = "test.db"
		}
	}
	logFlags := qb.LDefault
	if v, _ := os.LookupEnv("XBUS_TEST_SQL_LOGQUERY"); v != "" {
		logFlags |= qb.LQuery
	}
	if v, _ := os.LookupEnv("XBUS_TEST_SQL_LOGBINDINGS"); v != "" {
		logFlags |= qb.LBindings
	}
	tdb := TestDB{
		Driver: driver,
		Dsn:    dsn,
		Logger: testutils.GetTestLogger(t),
		t:      t,
	}

	if !func() bool {
		_, db := tdb.GetModelDB(1)
		defer db.Close()

		var allGood bool

		defer func() {
			if !allGood {
				tdb.Close()
			}
		}()
		if driver == "postgres" {
			lock := testutils.NewPortLock(1764)
			err := lock.Lock()
			if err != nil {
				t.Errorf("Cannot get lock: %s", err)
				return false
			}
			tdb.lock = lock
		}
		if droptables {
			tables := db.Metadata.GetQbMetadata().Tables()
			for i := len(tables) - 1; i >= 0; i-- {
				table := tables[i]
				drop := table.Drop(db.Engine.Dialect())
				if _, err := db.Engine.DB().Exec(drop); err != nil {
					t.Log(err)
				}
			}
			if _, err := db.Engine.DB().DB.Exec("DROP TABLE account_session"); err != nil {
				t.Log(err)
			}
			if _, err := db.Engine.DB().DB.Exec("DROP TABLE job"); err != nil {
				t.Log(err)
			}
			if _, err := db.Engine.DB().DB.Exec("DROP TABLE passwords"); err != nil {
				t.Log(err)
			}
			if _, err := db.Engine.DB().DB.Exec("DROP TABLE schema_migrations"); err != nil {
				t.Log(err)
			}
		}
		return true
	}() {
		return nil
	}

	_, db := tdb.GetModelDB(1)
	defer db.Close()

	m, _, err := migration.New(db)
	if err != nil {
		t.Errorf("Failed to init migration engine: %s", err)
		return nil
	}
	defer m.Close()

	m.Log = migrateLogger{tdb.Logger}

	if err := m.Up(); err != nil {
		t.Error(err)
		return nil
	}

	return &tdb
}
