package storage

import (
	"testing"
	"time"

	"github.com/stretchr/testify/assert"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
)

// SessionStorageFactory is any factory of SessionStorage
type SessionStorageFactory func(bt testing.TB) (storage.SessionStorage, func())

// SessionStorageTests is a test suite for SessionStorage implementations
func SessionStorageTests(t *testing.T, factory SessionStorageFactory) {

	t.Run("SetAndGet", func(t *testing.T) {
		store, clean := factory(t)
		if store == nil {
			t.Fatal("No store returned by the factory")
		}
		defer clean()

		var session = storage.Session{
			SessionToken: api.SessionToken{
				Token:      "thetoken",
				ValidUntil: time.Date(2019, 8, 27, 20, 0, 0, 0, time.UTC),
			},
			AccountID: api.MustUUIDFromString("483b8532-a6f2-11e8-b206-13d021cd09cd"),
		}

		err := store.Set(session)
		if err != nil {
			t.Fatal(err)
		}

		s, err := store.Get(session.Token)
		if err != nil {
			t.Fatal(err)
		}

		if s == nil {
			t.Fatal("Get() returned (nil, nil)")
		}
		assert.Equal(t, session, *s)
	})

	t.Run("List", func(t *testing.T) {
		store, clean := factory(t)
		if store == nil {
			t.Fatal("No store returned by the factory")
		}
		defer clean()

		list, err := store.List()
		if err != nil {
			t.Fatal(err)
		}

		assert.Empty(t, list)

		var session = storage.Session{
			SessionToken: api.SessionToken{
				Token:      "thetoken",
				ValidUntil: time.Date(2019, 8, 27, 20, 0, 0, 0, time.UTC),
			},
			AccountID: api.MustUUIDFromString("483b8532-a6f2-11e8-b206-13d021cd09cd"),
		}

		if err := store.Set(session); err != nil {
			t.Fatal(err)
		}

		list, err = store.List()
		if err != nil {
			t.Fatal(err)
		}

		assert.Equal(t, []storage.Session{session}, list)
	})

	t.Run("Delete", func(t *testing.T) {
		store, clean := factory(t)
		if store == nil {
			t.Fatal("No store returned by the factory")
		}
		defer clean()

		var session = storage.Session{
			SessionToken: api.SessionToken{
				Token:      "thetoken",
				ValidUntil: time.Date(2019, 8, 27, 20, 0, 0, 0, time.UTC),
			},
			AccountID: api.MustUUIDFromString("483b8532-a6f2-11e8-b206-13d021cd09cd"),
		}

		if err := store.Delete(session.Token); err != nil {
			t.Fatal(err)
		}

		if err := store.Set(session); err != nil {
			t.Fatal(err)
		}

		s, err := store.Get(session.Token)
		if err != nil {
			t.Fatal(err)
		}

		if s == nil {
			t.Fatal("Get() returned (nil, nil)")
		}
		assert.Equal(t, session, *s)

		if err := store.Delete(session.Token); err != nil {
			t.Fatal(err)
		}

		s, err = store.Get(session.Token)
		if err != nil {
			t.Fatal(err)
		}
		assert.Nil(t, s)

		list, err := store.List()
		if err != nil {
			t.Fatal(err)
		}
		assert.Empty(t, list)
	})

	t.Run("DeleteOlderThan", func(t *testing.T) {
		store, clean := factory(t)
		if store == nil {
			t.Fatal("No store returned by the factory")
		}
		defer clean()

		var session1 = storage.Session{
			SessionToken: api.SessionToken{
				Token:      "thetoken1",
				ValidUntil: time.Date(2019, 8, 27, 20, 0, 0, 0, time.UTC),
			},
			AccountID: api.MustUUIDFromString("483b8532-a6f2-11e8-b206-13d021cd09cd"),
		}

		var session2 = storage.Session{
			SessionToken: api.SessionToken{
				Token:      "thetoken2",
				ValidUntil: time.Date(2019, 8, 27, 21, 0, 0, 0, time.UTC),
			},
			AccountID: api.MustUUIDFromString("483b8532-a6f2-11e8-b206-13d021cd09cd"),
		}

		if err := store.Set(session1); err != nil {
			t.Fatal(err)
		}

		if err := store.Set(session2); err != nil {
			t.Fatal(err)
		}

		list, err := store.List()
		if err != nil {
			t.Fatal(err)
		}

		assert.Equal(t, 2, len(list))
		assert.Contains(t, list, session1)
		assert.Contains(t, list, session2)

		if err := store.DeleteOlderThan(
			time.Date(2019, 8, 27, 20, 30, 0, 0, time.UTC),
		); err != nil {
			t.Fatal(err)
		}

		list, err = store.List()
		if err != nil {
			t.Fatal(err)
		}

		assert.Equal(t, 1, len(list))
		assert.Contains(t, list, session2)
	})
}
