package xbus

import (
	"sync"
	"time"

	api "bitbucket.org/orus-io/go-xbus/api"
)

// SendEnvelopeFunc ...
type SendEnvelopeFunc func(*api.Envelope) (api.EnvelopeAck_ReceptionStatus, error)

// NewEnvelopeSender creates a new envelope sender
func NewEnvelopeSender(
	eventTypes []string,
	maxSize int,
	sendEnvelope SendEnvelopeFunc,
) *EnvelopeSender {
	return NewEnvelopeSenderWithIDs(
		api.UUID{},
		eventTypes,
		make([]api.UUID, len(eventTypes)),
		maxSize,
		sendEnvelope,
	)
}

// NewEnvelopeSenderWithIDs returns a EnvelopeSender with forced envelope and
// event ids. If any ID is zero, an ID is forged
func NewEnvelopeSenderWithIDs(
	envelopeID api.UUID,
	eventTypes []string,
	eventIDs []api.UUID,
	maxSize int,
	sendEnvelope SendEnvelopeFunc,
) *EnvelopeSender {
	if envelopeID.IsZero() {
		var err error
		envelopeID, err = api.UUIDV4()
		if err != nil {
			panic(err)
		}
	}
	es := EnvelopeSender{
		sendEnvelope: sendEnvelope,
		itemQueues:   make([]ItemQueue, len(eventTypes)),
		eventIndexes: make(map[string]int),
		maxSize:      maxSize,
		maxItems:     1000,
		envelope: api.Envelope{
			ID: envelopeID,
		},
	}
	for i, eType := range eventTypes {
		eventID := eventIDs[i]
		if eventID.IsZero() {
			var err error
			eventID, err = api.UUIDV4()
			if err != nil {
				panic(err)
			}
		}
		es.envelope.EventIDs = append(es.envelope.EventIDs, eventID)
		es.envelope.Events = append(es.envelope.Events, api.Event{
			ID:       eventID,
			Type:     eType,
			Index:    0,
			Checksum: api.Checksum(0).Update(eventID[:]).Update([]byte(eType)),
		})
		es.eventIndexes[eType] = i
	}
	return &es
}

// EnvelopeSender streams items as one or several envelopes
// to a given Actor output.
type EnvelopeSender struct {
	sendEnvelope SendEnvelopeFunc
	eventIndexes map[string]int
	queuesSize   int
	itemQueues   []ItemQueue
	envelope     api.Envelope
	status       api.EnvelopeAck_ReceptionStatus
	maxSize      int
	maxItems     int

	minDelay       time.Duration
	keepAliveTimer *time.Timer

	mu sync.Mutex
}

// send emits a fragment
func (s *EnvelopeSender) send(fragment *api.Envelope) (api.EnvelopeAck_ReceptionStatus, error) {
	if s.keepAliveTimer != nil {
		s.keepAliveTimer.Reset(s.minDelay)
	}
	return s.sendEnvelope(fragment)
}

// SetMaxDelay set the maximum delay between two 'send' calls. If no data is
// available and empty fragment is sent.
func (s *EnvelopeSender) SetMaxDelay(delay time.Duration) {
	s.mu.Lock()
	defer s.mu.Unlock()

	s.minDelay = delay

	if delay == 0 && s.keepAliveTimer != nil {
		s.keepAliveTimer.Stop()
		s.keepAliveTimer = nil
		return
	}
	if s.keepAliveTimer == nil {
		s.keepAliveTimer = time.NewTimer(delay)
		go func(c <-chan time.Time) {
			for range c {
				s.Flush()
			}
		}(s.keepAliveTimer.C)
	} else {
		s.keepAliveTimer.Reset(delay)
	}
}

// GetID returns the ID of the envelope
func (s *EnvelopeSender) GetID() api.UUID {
	return s.envelope.ID
}

func (s *EnvelopeSender) addItems(index int, items []api.Item) error {
	s.itemQueues[index].Add(items...)
	event := &s.envelope.Events[index]
	event.ItemCount += uint64(len(items))
	for _, item := range items {
		event.Checksum = event.Checksum.Update(item)
		s.queuesSize += len(item)
	}
	return s.autoFlush()
}

// AddItems pushes items to the event of the given type in the current envelope
func (s *EnvelopeSender) AddItems(eventType string, items ...api.Item) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	index, ok := s.eventIndexes[eventType]
	if !ok {
		panic("Invalid eventType")
	}
	return s.addItems(index, items)
}

func (s *EnvelopeSender) nextFragment() (api.Envelope, bool) {
	env := api.Envelope{
		ID:       s.envelope.ID,
		EventIDs: s.envelope.EventIDs,
	}
	envSize := 0
	for i := range s.envelope.Events {
		event := &s.envelope.Events[i]
		event.Index++

		var items []api.Item
		if len(s.itemQueues[i]) > 0 {
			var size int
			items, size = s.itemQueues[i].RemoveNS(s.maxItems, s.maxSize-envSize)
			s.queuesSize -= size
			envSize += size
		}

		env.Events = append(env.Events, api.Event{
			ID:    event.ID,
			Type:  event.Type,
			Index: event.Index,
			Items: items,
		})
	}
	return env, s.queuesSize == 0
}

// autoFlush emit envelopes until queue sizes is lower that the cache limit
func (s *EnvelopeSender) autoFlush() error {
	biggestQueueLen := func() int {
		n := 0
		for _, q := range s.itemQueues {
			if len(q) > n {
				n = len(q)
			}
		}
		return n
	}
	for biggestQueueLen() > s.maxItems || s.queuesSize > s.maxSize {
		env, _ := s.nextFragment()
		status, err := s.send(&env)
		if err != nil {
			return err
		}
		s.status = status
	}
	return nil
}

// Flush forces the buffered items to be sent right away
func (s *EnvelopeSender) Flush() {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.status == api.EnvelopeAck_ACCEPTED ||
		s.status == api.EnvelopeAck_ERROR {
		return
	}
	for {
		env, last := s.nextFragment()
		status, err := s.send(&env)
		if err != nil {
			panic(err)
		}
		s.status = status
		if last {
			break
		}
	}
}

// Close closes the streams. Must be called to finalize the emission
func (s *EnvelopeSender) Close() error {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.keepAliveTimer != nil {
		s.keepAliveTimer.Stop()
		s.keepAliveTimer = nil
	}

	// make sure we only have on more envelope
	if err := s.autoFlush(); err != nil {
		return err
	}

	env, last := s.nextFragment()
	if !last {
		panic("Should be the last envelope, something is wrong")
	}
	var noItems bool
	for i, event := range s.envelope.Events {
		env.Events[i].ItemCount = event.ItemCount
		env.Events[i].Checksum = event.Checksum
		if event.ItemCount != 0 {
			noItems = false
		}
	}
	var (
		status api.EnvelopeAck_ReceptionStatus
		err    error
	)
	if noItems {
		status, err = s.send(nil)
	} else {
		status, err = s.send(&env)
	}
	if err != nil {
		return err
	}
	s.status = status
	return nil
}
